@@ -52,33 +52,137 @@ static int msg_fetch_header (CONTEXT* ctx, IMAP_HEADER* h, char* buf,
 static int msg_parse_fetch (IMAP_HEADER* h, char* s);
 static char* msg_parse_flags (IMAP_HEADER* h, char* s);
 
+static void imap_update_context (IMAP_DATA *idata, int oldmsgcount)
+{
+  CONTEXT *ctx;
+  HEADER *h;
+  int msgno;
+
+  ctx = idata->ctx;
+  if (!idata->uid_hash)
+    idata->uid_hash = int_hash_create (MAX (6 * ctx->msgcount / 5, 30), 0);
+
+  for (msgno = oldmsgcount; msgno < ctx->msgcount; msgno++)
+  {
+    h = ctx->hdrs[msgno];
+    int_hash_insert (idata->uid_hash, HEADER_DATA(h)->uid, h);
+  }
+}
+
+static void imap_alloc_msn_index (IMAP_DATA *idata, unsigned int msn_count)
+{
+  unsigned int new_size;
+
+  if (msn_count <= idata->msn_index_size)
+    return;
+
+  /* This is a conservative check to protect against a malicious imap
+   * server.  Most likely size_t is bigger than an unsigned int, but
+   * if msn_count is this big, we have a serious problem. */
+  if (msn_count >= (UINT_MAX / sizeof (HEADER *)))
+  {
+    mutt_error _("Integer overflow -- can't allocate memory.");
+    sleep (1);
+    mutt_exit (1);
+  }
+
+  /* Add a little padding, like mx_allloc_memory() */
+  new_size = msn_count + 25;
+
+  if (!idata->msn_index)
+    idata->msn_index = safe_calloc (new_size, sizeof (HEADER *));
+  else
+  {
+    safe_realloc (&idata->msn_index, sizeof (HEADER *) * new_size);
+    memset (idata->msn_index + idata->msn_index_size, 0,
+            sizeof (HEADER *) * (new_size - idata->msn_index_size));
+  }
+
+  idata->msn_index_size = new_size;
+}
+
+/* Generates a more complicated sequence set after using the header cache,
+ * in case there are missing MSNs in the middle.
+ *
+ * There is a suggested limit of 1000 bytes for an IMAP client request.
+ * Ideally, we would generate multiple requests if the number of ranges
+ * is too big, but for now just abort to using the whole range.
+ */
+static void imap_generate_seqset (BUFFER *b, IMAP_DATA *idata, unsigned int msn_begin,
+                                  unsigned int msn_end)
+{
+  int chunks = 0;
+  int state = 0;  /* 1: single msn, 2: range of msn */
+  unsigned int msn, range_begin, range_end;
+
+  for (msn = msn_begin; msn <= msn_end + 1; msn++)
+  {
+    if (msn <= msn_end && !idata->msn_index[msn-1])
+    {
+      switch (state)
+      {
+        case 1:            /* single: convert to a range */
+          state = 2;
+          /* fall through */
+        case 2:            /* extend range ending */
+          range_end = msn;
+          break;
+        default:
+          state = 1;
+          range_begin = msn;
+          break;
+      }
+    }
+    else if (state)
+    {
+      if (chunks++)
+        mutt_buffer_addch (b, ',');
+      if (chunks == 150)
+        break;
+
+      if (state == 1)
+        mutt_buffer_printf (b, "%u", range_begin);
+      else if (state == 2)
+        mutt_buffer_printf (b, "%u:%u", range_begin, range_end);
+      state = 0;
+    }
+  }
+
+  /* Too big.  Just query the whole range then. */
+  if (chunks == 150 || mutt_strlen (b->data) > 500)
+  {
+    b->dptr = b->data;
+    mutt_buffer_printf (b, "%u:%u", msn_begin, msn_end);
+  }
+}
+
 /* imap_read_headers:
  * Changed to read many headers instead of just one. It will return the
- * msgno of the last message read. It will return a value other than
- * msgend if mail comes in while downloading headers (in theory).
+ * msn of the last message read. It will return a value other than
+ * msn_end if mail comes in while downloading headers (in theory).
  */
-int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
+int imap_read_headers (IMAP_DATA* idata, unsigned int msn_begin, unsigned int msn_end)
 {
   CONTEXT* ctx;
   char *hdrreq = NULL;
   FILE *fp;
   char tempfile[_POSIX_PATH_MAX];
-  int msgno, idx = msgbegin - 1;
+  int msgno, idx;
   IMAP_HEADER h;
   IMAP_STATUS* status;
-  int rc, mfhrc, oldmsgcount;
-  int fetchlast = 0;
-  int maxuid = 0;
+  int rc, mfhrc = 0, oldmsgcount;
+  int fetch_msn_end = 0;
+  unsigned int maxuid = 0;
   static const char * const want_headers = "DATE FROM SUBJECT TO CC MESSAGE-ID REFERENCES CONTENT-TYPE CONTENT-DESCRIPTION IN-REPLY-TO REPLY-TO LINES LIST-POST X-LABEL";
   progress_t progress;
   int retval = -1;
+  int evalhc = 0;
 
 #if USE_HCACHE
   char buf[LONG_STRING];
   unsigned int *uid_validity = NULL;
   unsigned int *puidnext = NULL;
   unsigned int uidnext = 0;
-  int evalhc = 0;
 #endif /* USE_HCACHE */
 
   ctx = idata->ctx;
@@ -112,9 +216,11 @@ int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
   unlink (tempfile);
 
   /* make sure context has room to hold the mailbox */
-  while ((msgend) >= idata->ctx->hdrmax)
-    mx_alloc_memory (idata->ctx);
+  while (msn_end > ctx->hdrmax)
+    mx_alloc_memory (ctx);
+  imap_alloc_msn_index (idata, msn_end);
 
+  idx = ctx->msgcount;
   oldmsgcount = ctx->msgcount;
   idata->reopen &= ~(IMAP_REOPEN_ALLOW|IMAP_NEWMAIL_PENDING);
   idata->newMailCount = 0;
@@ -122,25 +228,25 @@ int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
 #if USE_HCACHE
   idata->hcache = imap_hcache_open (idata, NULL);
 
-  if (idata->hcache && !msgbegin)
+  if (idata->hcache && (msn_begin == 1))
   {
     uid_validity = mutt_hcache_fetch_raw (idata->hcache, "/UIDVALIDITY", imap_hcache_keylen);
     puidnext = mutt_hcache_fetch_raw (idata->hcache, "/UIDNEXT", imap_hcache_keylen);
     if (puidnext)
     {
       uidnext = *puidnext;
-      FREE (&puidnext);
+      mutt_hcache_free ((void **)&puidnext);
     }
     if (uid_validity && uidnext && *uid_validity == idata->uid_validity)
       evalhc = 1;
-    FREE (&uid_validity);
+    mutt_hcache_free ((void **)&uid_validity);
   }
   if (evalhc)
   {
     /* L10N:
        Comparing the cached data with the IMAP server's data */
     mutt_progress_init (&progress, _("Evaluating cache..."),
-			M_PROGRESS_MSG, ReadInc, msgend + 1);
+			MUTT_PROGRESS_MSG, ReadInc, msn_end);
 
     snprintf (buf, sizeof (buf),
       "UID FETCH 1:%u (UID FLAGS)", uidnext - 1);
@@ -148,47 +254,48 @@ int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
     imap_cmd_start (idata, buf);
 
     rc = IMAP_CMD_CONTINUE;
-    for (msgno = msgbegin; rc == IMAP_CMD_CONTINUE; msgno++)
+    for (msgno = 1; rc == IMAP_CMD_CONTINUE; msgno++)
     {
-      mutt_progress_update (&progress, msgno + 1, -1);
+      mutt_progress_update (&progress, msgno, -1);
 
       memset (&h, 0, sizeof (h));
       h.data = safe_calloc (1, sizeof (IMAP_HEADER_DATA));
       do
       {
-        mfhrc = 0;
-
         rc = imap_cmd_step (idata);
         if (rc != IMAP_CMD_CONTINUE)
-	{
-	  imap_free_header_data (&h.data);
           break;
-	}
 
-        /* hole in the header cache */
-        if (!evalhc)
+        if ((mfhrc = msg_fetch_header (ctx, &h, idata->buf, NULL)) < 0)
           continue;
 
-        if ((mfhrc = msg_fetch_header (ctx, &h, idata->buf, NULL)) == -1)
-          continue;
-        else if (mfhrc < 0)
-	{
-	  imap_free_header_data (&h.data);
-          break;
-	}
-
         if (!h.data->uid)
         {
           dprint (2, (debugfile, "imap_read_headers: skipping hcache FETCH "
-                      "response for unknown message number %d\n", h.sid));
-          mfhrc = -1;
+                      "response for message number %d missing a UID\n", h.data->msn));
+          continue;
+        }
+
+        if (h.data->msn < 1 || h.data->msn > msn_end)
+        {
+          dprint (1, (debugfile, "imap_read_headers: skipping hcache FETCH "
+                      "response for unknown message number %d\n", h.data->msn));
+          continue;
+        }
+
+        if (idata->msn_index[h.data->msn - 1])
+        {
+          dprint (2, (debugfile, "imap_read_headers: skipping hcache FETCH "
+                      "for duplicate message %d\n", h.data->msn));
           continue;
         }
 
-        idx++;
         ctx->hdrs[idx] = imap_hcache_get (idata, h.data->uid);
         if (ctx->hdrs[idx])
         {
+          idata->max_msn = MAX (idata->max_msn, h.data->msn);
+          idata->msn_index[h.data->msn - 1] = ctx->hdrs[idx];
+
   	  ctx->hdrs[idx]->index = idx;
   	  /* messages which have not been expunged are ACTIVE (borrowed from mh
   	   * folders) */
@@ -204,158 +311,178 @@ int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
 
           ctx->msgcount++;
           ctx->size += ctx->hdrs[idx]->content->length;
-        }
-	else
-        {
-	  /* bad header in the cache, we'll have to refetch. */
-          dprint (3, (debugfile, "bad cache entry at %d, giving up\n", h.sid - 1));
-          imap_free_header_data(&h.data);
-          evalhc = 0;
-          idx--;
+
+          h.data = NULL;
+          idx++;
         }
       }
-      while (rc != IMAP_CMD_OK && mfhrc == -1);
-      if (rc == IMAP_CMD_OK)
-        break;
+      while (mfhrc == -1);
+
+      imap_free_header_data (&h.data);
+
       if ((mfhrc < -1) || ((rc != IMAP_CMD_CONTINUE) && (rc != IMAP_CMD_OK)))
       {
-        imap_free_header_data (&h.data);
         imap_hcache_close (idata);
 	goto error_out_1;
       }
     }
-    /* could also look for first null header in case hcache is holey */
-    msgbegin = ctx->msgcount;
+
+    /* Look for the first empty MSN and start there */
+    while (msn_begin <= msn_end)
+    {
+      if (!idata->msn_index[msn_begin -1])
+        break;
+      msn_begin++;
+    }
   }
 #endif /* USE_HCACHE */
 
   mutt_progress_init (&progress, _("Fetching message headers..."),
-		      M_PROGRESS_MSG, ReadInc, msgend + 1);
+		      MUTT_PROGRESS_MSG, ReadInc, msn_end);
 
-  for (msgno = msgbegin; msgno <= msgend ; msgno++)
+  while (msn_begin <= msn_end && fetch_msn_end < msn_end)
   {
-    mutt_progress_update (&progress, msgno + 1, -1);
+    char *cmd;
+    BUFFER *b;
 
-    /* we may get notification of new mail while fetching headers */
-    if (msgno + 1 > fetchlast)
+    b = mutt_buffer_new ();
+    if (evalhc)
     {
-      char *cmd;
-
-      fetchlast = msgend + 1;
-      safe_asprintf (&cmd, "FETCH %d:%d (UID FLAGS INTERNALDATE RFC822.SIZE %s)",
-                     msgno + 1, fetchlast, hdrreq);
-      imap_cmd_start (idata, cmd);
-      FREE (&cmd);
+      /* In case there are holes in the header cache. */
+      evalhc = 0;
+      imap_generate_seqset (b, idata, msn_begin, msn_end);
     }
+    else
+      mutt_buffer_printf (b, "%u:%u", msn_begin, msn_end);
 
-    rewind (fp);
-    memset (&h, 0, sizeof (h));
-    h.data = safe_calloc (1, sizeof (IMAP_HEADER_DATA));
+    fetch_msn_end = msn_end;
+    safe_asprintf (&cmd, "FETCH %s (UID FLAGS INTERNALDATE RFC822.SIZE %s)",
+                   b->data, hdrreq);
+    imap_cmd_start (idata, cmd);
+    FREE (&cmd);
+    mutt_buffer_free (&b);
 
-    /* this DO loop does two things:
-     * 1. handles untagged messages, so we can try again on the same msg
-     * 2. fetches the tagged response at the end of the last message.
-     */
-    do
+    rc = IMAP_CMD_CONTINUE;
+    for (msgno = msn_begin; rc == IMAP_CMD_CONTINUE; msgno++)
     {
-      mfhrc = 0;
-
-      rc = imap_cmd_step (idata);
-      if (rc != IMAP_CMD_CONTINUE)
-	break;
-
-      if ((mfhrc = msg_fetch_header (ctx, &h, idata->buf, fp)) == -1)
-	continue;
-      else if (mfhrc < 0)
-	break;
-
-      if (!ftello (fp))
-      {
-        dprint (2, (debugfile, "msg_fetch_header: ignoring fetch response with no body\n"));
-        mfhrc = -1;
-        msgend--;
-        continue;
-      }
-
-      /* make sure we don't get remnants from older larger message headers */
-      fputs ("\n\n", fp);
-
-      idx++;
-      if (idx > msgend)
-      {
-        dprint (1, (debugfile, "imap_read_headers: skipping FETCH response for "
-                    "unknown message number %d\n", h.sid));
-        mfhrc = -1;
-        idx--;
-        continue;
-      }
-      /* May receive FLAGS updates in a separate untagged response (#2935) */
-      if (idx < ctx->msgcount)
-      {
-	dprint (2, (debugfile, "imap_read_headers: message %d is not new\n",
-		    h.sid));
-        idx--;
-	continue;
-      }
-
-      ctx->hdrs[idx] = mutt_new_header ();
-
-      ctx->hdrs[idx]->index = h.sid - 1;
-      /* messages which have not been expunged are ACTIVE (borrowed from mh
-       * folders) */
-      ctx->hdrs[idx]->active = 1;
-      ctx->hdrs[idx]->read = h.data->read;
-      ctx->hdrs[idx]->old = h.data->old;
-      ctx->hdrs[idx]->deleted = h.data->deleted;
-      ctx->hdrs[idx]->flagged = h.data->flagged;
-      ctx->hdrs[idx]->replied = h.data->replied;
-      ctx->hdrs[idx]->changed = h.data->changed;
-      ctx->hdrs[idx]->received = h.received;
-      ctx->hdrs[idx]->data = (void *) (h.data);
-
-      if (maxuid < h.data->uid)
-        maxuid = h.data->uid;
+      mutt_progress_update (&progress, msgno, -1);
 
       rewind (fp);
-      /* NOTE: if Date: header is missing, mutt_read_rfc822_header depends
-       *   on h.received being set */
-      ctx->hdrs[idx]->env = mutt_read_rfc822_header (fp, ctx->hdrs[idx],
-        0, 0);
-      /* content built as a side-effect of mutt_read_rfc822_header */
-      ctx->hdrs[idx]->content->length = h.content_length;
-      ctx->size += h.content_length;
+      memset (&h, 0, sizeof (h));
+      h.data = safe_calloc (1, sizeof (IMAP_HEADER_DATA));
+
+      /* this DO loop does two things:
+       * 1. handles untagged messages, so we can try again on the same msg
+       * 2. fetches the tagged response at the end of the last message.
+       */
+      do
+      {
+        rc = imap_cmd_step (idata);
+        if (rc != IMAP_CMD_CONTINUE)
+          break;
+
+        if ((mfhrc = msg_fetch_header (ctx, &h, idata->buf, fp)) < 0)
+          continue;
+
+        if (!ftello (fp))
+        {
+          dprint (2, (debugfile, "msg_fetch_header: ignoring fetch response with no body\n"));
+          continue;
+        }
+
+        /* make sure we don't get remnants from older larger message headers */
+        fputs ("\n\n", fp);
+
+        if (h.data->msn < 1 || h.data->msn > fetch_msn_end)
+        {
+          dprint (1, (debugfile, "imap_read_headers: skipping FETCH response for "
+                      "unknown message number %d\n", h.data->msn));
+          continue;
+        }
+
+        /* May receive FLAGS updates in a separate untagged response (#2935) */
+        if (idata->msn_index[h.data->msn - 1])
+        {
+          dprint (2, (debugfile, "imap_read_headers: skipping FETCH response for "
+                      "duplicate message %d\n", h.data->msn));
+          continue;
+        }
+
+        ctx->hdrs[idx] = mutt_new_header ();
+
+        idata->max_msn = MAX (idata->max_msn, h.data->msn);
+        idata->msn_index[h.data->msn - 1] = ctx->hdrs[idx];
+
+        ctx->hdrs[idx]->index = idx;
+        /* messages which have not been expunged are ACTIVE (borrowed from mh
+         * folders) */
+        ctx->hdrs[idx]->active = 1;
+        ctx->hdrs[idx]->read = h.data->read;
+        ctx->hdrs[idx]->old = h.data->old;
+        ctx->hdrs[idx]->deleted = h.data->deleted;
+        ctx->hdrs[idx]->flagged = h.data->flagged;
+        ctx->hdrs[idx]->replied = h.data->replied;
+        ctx->hdrs[idx]->changed = h.data->changed;
+        ctx->hdrs[idx]->received = h.received;
+        ctx->hdrs[idx]->data = (void *) (h.data);
+
+        if (maxuid < h.data->uid)
+          maxuid = h.data->uid;
+
+        rewind (fp);
+        /* NOTE: if Date: header is missing, mutt_read_rfc822_header depends
+         *   on h.received being set */
+        ctx->hdrs[idx]->env = mutt_read_rfc822_header (fp, ctx->hdrs[idx],
+                                                       0, 0);
+        /* content built as a side-effect of mutt_read_rfc822_header */
+        ctx->hdrs[idx]->content->length = h.content_length;
+        ctx->size += h.content_length;
 
 #if USE_HCACHE
-      imap_hcache_put (idata, ctx->hdrs[idx]);
+        imap_hcache_put (idata, ctx->hdrs[idx]);
 #endif /* USE_HCACHE */
 
-      ctx->msgcount++;
-    }
-    while ((rc != IMAP_CMD_OK) && ((mfhrc == -1) ||
-      ((msgno + 1) >= fetchlast)));
+        ctx->msgcount++;
+
+        h.data = NULL;
+        idx++;
+      }
+      while (mfhrc == -1);
 
-    if ((mfhrc < -1) || ((rc != IMAP_CMD_CONTINUE) && (rc != IMAP_CMD_OK)))
-    {
       imap_free_header_data (&h.data);
+
+      if ((mfhrc < -1) || ((rc != IMAP_CMD_CONTINUE) && (rc != IMAP_CMD_OK)))
+      {
 #if USE_HCACHE
-      imap_hcache_close (idata);
+        imap_hcache_close (idata);
 #endif
-      goto error_out_1;
+        goto error_out_1;
+      }
     }
 
-    /* in case we get new mail while fetching the headers */
+    /* In case we get new mail while fetching the headers.
+     *
+     * Note: The RFC says we shouldn't get any EXPUNGE responses in the
+     * middle of a FETCH.  But just to be cautious, use the current state
+     * of max_msn, not fetch_msn_end to set the next start range.
+     */
     if (idata->reopen & IMAP_NEWMAIL_PENDING)
     {
-      msgend = idata->newMailCount - 1;
-      while ((msgend) >= ctx->hdrmax)
-	mx_alloc_memory (ctx);
+      /* update to the last value we actually pulled down */
+      fetch_msn_end = idata->max_msn;
+      msn_begin = idata->max_msn + 1;
+      msn_end = idata->newMailCount;
+      while (msn_end > ctx->hdrmax)
+        mx_alloc_memory (ctx);
+      imap_alloc_msn_index (idata, msn_end);
       idata->reopen &= ~IMAP_NEWMAIL_PENDING;
       idata->newMailCount = 0;
     }
   }
 
-  if (maxuid && (status = imap_mboxcache_get (idata, idata->mailbox, 0)))
-  status->uidnext = maxuid + 1;
+  if (maxuid && (status = imap_mboxcache_get (idata, idata->mailbox, 0)) &&
+      (status->uidnext < maxuid + 1))
+    status->uidnext = maxuid + 1;
 
 #if USE_HCACHE
   mutt_hcache_store_raw (idata->hcache, "/UIDVALIDITY", &idata->uid_validity,
@@ -374,13 +501,16 @@ int imap_read_headers (IMAP_DATA* idata, int msgbegin, int msgend)
 
   if (ctx->msgcount > oldmsgcount)
   {
+    /* TODO: it's not clear to me why we are calling mx_alloc_memory
+     *       yet again. */
     mx_alloc_memory(ctx);
     mx_update_context (ctx, ctx->msgcount - oldmsgcount);
+    imap_update_context (idata, oldmsgcount);
   }
 
   idata->reopen |= IMAP_REOPEN_ALLOW;
 
-  retval = msgend;
+  retval = msn_end;
 
 error_out_1:
   safe_fclose (&fp);
@@ -391,7 +521,7 @@ error_out_0:
   return retval;
 }
 
-int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
+int imap_fetch_message (CONTEXT *ctx, MESSAGE *msg, int msgno)
 {
   IMAP_DATA* idata;
   HEADER* h;
@@ -399,9 +529,9 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
   char buf[LONG_STRING];
   char path[_POSIX_PATH_MAX];
   char *pc;
-  long bytes;
+  unsigned int bytes;
   progress_t progressbar;
-  int uid;
+  unsigned int uid;
   int cacheno;
   IMAP_CACHE *cache;
   int read;
@@ -409,6 +539,7 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
   /* Sam's weird courier server returns an OK response even when FETCH
    * fails. Thanks Sam. */
   short fetched = 0;
+  int output_progress;
 
   idata = (IMAP_DATA*) ctx->data;
   h = ctx->hdrs[msgno];
@@ -439,7 +570,10 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
     }
   }
 
-  if (!isendwin())
+  /* This function is called in a few places after endwin()
+   * e.g. _mutt_pipe_message(). */
+  output_progress = !isendwin ();
+  if (output_progress)
     mutt_message _("Fetching message...");
 
   if (!(msg->fp = msg_cache_put (idata, h)))
@@ -484,7 +618,8 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
 	if (ascii_strncasecmp ("UID", pc, 3) == 0)
 	{
 	  pc = imap_next_word (pc);
-	  uid = atoi (pc);
+	  if (mutt_atoui (pc, &uid) < 0)
+            goto bail;
 	  if (uid != HEADER_DATA(h)->uid)
 	    mutt_error (_("The message index is incorrect. Try reopening the mailbox."));
 	}
@@ -497,9 +632,13 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
 	    imap_error ("imap_fetch_message()", buf);
 	    goto bail;
 	  }
-	  mutt_progress_init (&progressbar, _("Fetching message..."),
-			      M_PROGRESS_SIZE, NetInc, bytes);
-	  if (imap_read_literal (msg->fp, idata, bytes, &progressbar) < 0)
+          if (output_progress)
+          {
+            mutt_progress_init (&progressbar, _("Fetching message..."),
+                                MUTT_PROGRESS_SIZE, NetInc, bytes);
+          }
+	  if (imap_read_literal (msg->fp, idata, bytes,
+                                 output_progress ? &progressbar : NULL) < 0)
 	    goto bail;
 	  /* pick up trailing line */
 	  if ((rc = imap_cmd_step (idata)) != IMAP_CMD_CONTINUE)
@@ -514,7 +653,7 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
 	 * incrementally update flags later, this won't stop us syncing */
 	else if ((ascii_strncasecmp ("FLAGS", pc, 5) == 0) && !h->changed)
 	{
-	  if ((pc = imap_set_flags (idata, h, pc)) == NULL)
+	  if ((pc = imap_set_flags (idata, h, pc, NULL)) == NULL)
 	    goto bail;
 	}
       }
@@ -559,7 +698,7 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
   if (read != h->read)
   {
     h->read = read;
-    mutt_set_flag (ctx, h, M_NEW, read);
+    mutt_set_flag (ctx, h, MUTT_NEW, read);
   }
 
   h->lines = 0;
@@ -595,6 +734,21 @@ bail:
   return -1;
 }
 
+int imap_close_message (CONTEXT *ctx, MESSAGE *msg)
+{
+  return safe_fclose (&msg->fp);
+}
+
+int imap_commit_message (CONTEXT *ctx, MESSAGE *msg)
+{
+  int r = safe_fclose (&msg->fp);
+
+  if (r)
+    return r;
+
+  return imap_append_message (ctx, msg);
+}
+
 int imap_append_message (CONTEXT *ctx, MESSAGE *msg)
 {
   IMAP_DATA* idata;
@@ -642,7 +796,7 @@ int imap_append_message (CONTEXT *ctx, MESSAGE *msg)
   rewind (fp);
 
   mutt_progress_init (&progressbar, _("Uploading message..."),
-		      M_PROGRESS_SIZE, NetInc, len);
+		      MUTT_PROGRESS_SIZE, NetInc, len);
 
   imap_munge_mbox_name (idata, mbox, sizeof (mbox), mailbox);
   imap_make_date (internaldate, msg->received);
@@ -747,7 +901,7 @@ int imap_copy_messages (CONTEXT* ctx, HEADER* h, char* dest, int delete)
   int rc;
   int n;
   IMAP_MBOX mx;
-  int err_continue = M_NO;
+  int err_continue = MUTT_NO;
   int triedcreate = 0;
 
   idata = (IMAP_DATA*) ctx->data;
@@ -800,7 +954,7 @@ int imap_copy_messages (CONTEXT* ctx, HEADER* h, char* dest, int delete)
         if (ctx->hdrs[n]->tagged && ctx->hdrs[n]->active &&
             ctx->hdrs[n]->changed)
         {
-          rc = imap_sync_message (idata, ctx->hdrs[n], &sync_cmd, &err_continue);
+          rc = imap_sync_message_for_copy (idata, ctx->hdrs[n], &sync_cmd, &err_continue);
           if (rc < 0)
           {
             dprint (1, (debugfile, "imap_copy_messages: could not sync\n"));
@@ -809,7 +963,7 @@ int imap_copy_messages (CONTEXT* ctx, HEADER* h, char* dest, int delete)
         }
       }
 
-      rc = imap_exec_msgset (idata, "UID COPY", mmbox, M_TAG, 0, 0);
+      rc = imap_exec_msgset (idata, "UID COPY", mmbox, MUTT_TAG, 0, 0);
       if (!rc)
       {
         dprint (1, (debugfile, "imap_copy_messages: No messages tagged\n"));
@@ -831,7 +985,7 @@ int imap_copy_messages (CONTEXT* ctx, HEADER* h, char* dest, int delete)
 
       if (h->active && h->changed)
       {
-        rc = imap_sync_message (idata, h, &sync_cmd, &err_continue);
+        rc = imap_sync_message_for_copy (idata, h, &sync_cmd, &err_continue);
         if (rc < 0)
         {
           dprint (1, (debugfile, "imap_copy_messages: could not sync\n"));
@@ -885,16 +1039,18 @@ int imap_copy_messages (CONTEXT* ctx, HEADER* h, char* dest, int delete)
       {
         if (ctx->hdrs[n]->tagged)
         {
-          mutt_set_flag (ctx, ctx->hdrs[n], M_DELETE, 1);
+          mutt_set_flag (ctx, ctx->hdrs[n], MUTT_DELETE, 1);
+          mutt_set_flag (ctx, ctx->hdrs[n], MUTT_PURGE, 1);
           if (option (OPTDELETEUNTAG))
-            mutt_set_flag (ctx, ctx->hdrs[n], M_TAG, 0);
+            mutt_set_flag (ctx, ctx->hdrs[n], MUTT_TAG, 0);
         }
       }
     else
     {
-      mutt_set_flag (ctx, h, M_DELETE, 1);
+      mutt_set_flag (ctx, h, MUTT_DELETE, 1);
+      mutt_set_flag (ctx, h, MUTT_PURGE, 1);
       if (option (OPTDELETEUNTAG))
-        mutt_set_flag (ctx, h, M_TAG, 0);
+        mutt_set_flag (ctx, h, MUTT_TAG, 0);
     }
   }
 
@@ -973,24 +1129,17 @@ int imap_cache_del (IMAP_DATA* idata, HEADER* h)
 
 static int msg_cache_clean_cb (const char* id, body_cache_t* bcache, void* data)
 {
-  unsigned int uv, uid, n;
+  unsigned int uv, uid;
   IMAP_DATA* idata = (IMAP_DATA*)data;
 
   if (sscanf (id, "%u-%u", &uv, &uid) != 2)
     return 0;
 
   /* bad UID */
-  if (uv != idata->uid_validity)
+  if (uv != idata->uid_validity ||
+      !int_hash_find (idata->uid_hash, uid))
     mutt_bcache_del (bcache, id);
 
-  /* TODO: presort UIDs, walk in order */
-  for (n = 0; n < idata->ctx->msgcount; n++)
-  {
-    if (uid == HEADER_DATA(idata->ctx->hdrs[n])->uid)
-      return 0;
-  }
-  mutt_bcache_del (bcache, id);
-
   return 0;
 }
 
@@ -1035,20 +1184,56 @@ void imap_free_header_data (IMAP_HEADER_DATA** data)
   }
 }
 
+/* Sets server_changes to 1 if a change to a flag is made, or in the
+ * case of local_changes, if a change to a flag _would_ have been
+ * made. */
+static void imap_set_changed_flag (CONTEXT *ctx, HEADER *h, int local_changes,
+                                   int *server_changes, int flag_name, int old_hd_flag,
+                                   int new_hd_flag, int h_flag)
+{
+  /* If there are local_changes, we only want to note if the server
+   * flags have changed, so we can set a reopen flag in
+   * cmd_parse_fetch().  We don't want to count a local modification
+   * to the header flag as a "change".
+   */
+  if ((old_hd_flag != new_hd_flag) || (!local_changes))
+  {
+    if (new_hd_flag != h_flag)
+    {
+      if (server_changes)
+        *server_changes = 1;
+
+      /* Local changes have priority */
+      if (!local_changes)
+        mutt_set_flag (ctx, h, flag_name, new_hd_flag);
+    }
+  }
+}
+
 /* imap_set_flags: fill out the message header according to the flags from
- *   the server. Expects a flags line of the form "FLAGS (flag flag ...)" */
-char* imap_set_flags (IMAP_DATA* idata, HEADER* h, char* s)
+ * the server. Expects a flags line of the form "FLAGS (flag flag ...)"
+ *
+ * Sets server_changes to 1 if a change to a flag is made, or in the
+ * case of h->changed, if a change to a flag _would_ have been
+ * made. */
+char* imap_set_flags (IMAP_DATA* idata, HEADER* h, char* s, int *server_changes)
 {
   CONTEXT* ctx = idata->ctx;
   IMAP_HEADER newh;
+  IMAP_HEADER_DATA old_hd;
   IMAP_HEADER_DATA* hd;
   unsigned char readonly;
+  int local_changes;
+
+  local_changes = h->changed;
 
   memset (&newh, 0, sizeof (newh));
   hd = h->data;
   newh.data = hd;
 
-  dprint (2, (debugfile, "imap_fetch_message: parsing FLAGS\n"));
+  memcpy (&old_hd, hd, sizeof(old_hd));
+
+  dprint (2, (debugfile, "imap_set_flags: parsing FLAGS\n"));
   if ((s = msg_parse_flags (&newh, s)) == NULL)
     return NULL;
 
@@ -1059,16 +1244,24 @@ char* imap_set_flags (IMAP_DATA* idata, HEADER* h, char* s)
   readonly = ctx->readonly;
   ctx->readonly = 0;
 
-  mutt_set_flag (ctx, h, M_NEW, !(hd->read || hd->old));
-  mutt_set_flag (ctx, h, M_OLD, hd->old);
-  mutt_set_flag (ctx, h, M_READ, hd->read);
-  mutt_set_flag (ctx, h, M_DELETE, hd->deleted);
-  mutt_set_flag (ctx, h, M_FLAG, hd->flagged);
-  mutt_set_flag (ctx, h, M_REPLIED, hd->replied);
+  /* This is redundant with the following two checks. Removing:
+   * mutt_set_flag (ctx, h, MUTT_NEW, !(hd->read || hd->old));
+   */
+  imap_set_changed_flag (ctx, h, local_changes, server_changes,
+                         MUTT_OLD, old_hd.old, hd->old, h->old);
+  imap_set_changed_flag (ctx, h, local_changes, server_changes,
+                         MUTT_READ, old_hd.read, hd->read, h->read);
+  imap_set_changed_flag (ctx, h, local_changes, server_changes,
+                         MUTT_DELETE, old_hd.deleted, hd->deleted, h->deleted);
+  imap_set_changed_flag (ctx, h, local_changes, server_changes,
+                         MUTT_FLAG, old_hd.flagged, hd->flagged, h->flagged);
+  imap_set_changed_flag (ctx, h, local_changes, server_changes,
+                         MUTT_REPLIED, old_hd.replied, hd->replied, h->replied);
 
   /* this message is now definitively *not* changed (mutt_set_flag
    * marks things changed as a side-effect) */
-  h->changed = 0;
+  if (!local_changes)
+    h->changed = 0;
   ctx->changed &= ~readonly;
   ctx->readonly = readonly;
 
@@ -1085,8 +1278,9 @@ char* imap_set_flags (IMAP_DATA* idata, HEADER* h, char* s)
 static int msg_fetch_header (CONTEXT* ctx, IMAP_HEADER* h, char* buf, FILE* fp)
 {
   IMAP_DATA* idata;
-  long bytes;
+  unsigned int bytes;
   int rc = -1; /* default now is that string isn't FETCH response*/
+  int parse_rc;
 
   idata = (IMAP_DATA*) ctx->data;
 
@@ -1095,7 +1289,8 @@ static int msg_fetch_header (CONTEXT* ctx, IMAP_HEADER* h, char* buf, FILE* fp)
 
   /* skip to message number */
   buf = imap_next_word (buf);
-  h->sid = atoi (buf);
+  if (mutt_atoui (buf, &h->data->msn) < 0)
+    return rc;
 
   /* find FETCH tag */
   buf = imap_next_word (buf);
@@ -1109,7 +1304,10 @@ static int msg_fetch_header (CONTEXT* ctx, IMAP_HEADER* h, char* buf, FILE* fp)
 
   /* FIXME: current implementation - call msg_parse_fetch - if it returns -2,
    *   read header lines and call it again. Silly. */
-  if ((rc = msg_parse_fetch (h, buf)) != -2 || !fp)
+  parse_rc = msg_parse_fetch (h, buf);
+  if (!parse_rc)
+    return 0;
+  if (parse_rc != -2 || !fp)
     return rc;
 
   if (imap_get_literal_count (buf, &bytes) == 0)
@@ -1136,11 +1334,18 @@ static int msg_fetch_header (CONTEXT* ctx, IMAP_HEADER* h, char* buf, FILE* fp)
   return rc;
 }
 
-/* msg_parse_fetch: handle headers returned from header fetch */
+/* msg_parse_fetch: handle headers returned from header fetch.
+ * Returns:
+ *   0 on success
+ *   -1 if the string is corrupted
+ *   -2 if the fetch contains a body or header lines
+ *      that still need to be parsed.
+ */
 static int msg_parse_fetch (IMAP_HEADER *h, char *s)
 {
   char tmp[SHORT_STRING];
   char *ptmp;
+  size_t dlen;
 
   if (!s)
     return -1;
@@ -1158,7 +1363,8 @@ static int msg_parse_fetch (IMAP_HEADER *h, char *s)
     {
       s += 3;
       SKIPWS (s);
-      h->data->uid = (unsigned int) atoi (s);
+      if (mutt_atoui (s, &h->data->uid) < 0)
+        return -1;
 
       s = imap_next_word (s);
     }
@@ -1173,8 +1379,12 @@ static int msg_parse_fetch (IMAP_HEADER *h, char *s)
       }
       s++;
       ptmp = tmp;
-      while (*s && *s != '\"')
+      dlen = sizeof(tmp) - 1;
+      while (*s && *s != '\"' && dlen)
+      {
         *ptmp++ = *s++;
+        dlen--;
+      }
       if (*s != '\"')
         return -1;
       s++; /* skip past the trailing " */
@@ -1186,10 +1396,15 @@ static int msg_parse_fetch (IMAP_HEADER *h, char *s)
       s += 11;
       SKIPWS (s);
       ptmp = tmp;
-      while (isdigit ((unsigned char) *s))
+      dlen = sizeof(tmp) - 1;
+      while (isdigit ((unsigned char) *s) && dlen)
+      {
         *ptmp++ = *s++;
+        dlen--;
+      }
       *ptmp = 0;
-      h->content_length = atoi (tmp);
+      if (mutt_atol (tmp, &h->content_length) < 0)
+        return -1;
     }
     else if (!ascii_strncasecmp ("BODY", s, 4) ||
       !ascii_strncasecmp ("RFC822.HEADER", s, 13))
