@@ -23,7 +23,7 @@
 #include "config.h"
 
 #include "mutt.h"
-#include "mx.h"	/* for M_IMAP */
+#include "mx.h"	/* for MUTT_IMAP */
 #include "url.h"
 #include "imap_private.h"
 #ifdef USE_HCACHE
@@ -61,7 +61,7 @@ int imap_expand_path (char* path, size_t len)
   if (imap_parse_path (path, &mx) < 0)
     return -1;
 
-  idata = imap_conn_find (&mx.account, M_IMAP_CONN_NONEW);
+  idata = imap_conn_find (&mx.account, MUTT_IMAP_CONN_NONEW);
   mutt_account_tourl (&mx.account, &url);
   imap_fix_path (idata, mx.mbox, fixedpath, sizeof (fixedpath));
   url.path = fixedpath;
@@ -84,6 +84,7 @@ header_cache_t* imap_hcache_open (IMAP_DATA* idata, const char* path)
   ciss_url_t url;
   char cachepath[LONG_STRING];
   char mbox[LONG_STRING];
+  size_t len;
 
   if (path)
     imap_cachepath (idata, path, mbox, sizeof (mbox));
@@ -96,6 +97,12 @@ header_cache_t* imap_hcache_open (IMAP_DATA* idata, const char* path)
     FREE (&mx.mbox);
   }
 
+  if (strstr(mbox, "/../") || (strcmp(mbox, "..") == 0) || (strncmp(mbox, "../", 3) == 0))
+    return NULL;
+  len = strlen(mbox);
+  if ((len > 3) && (strcmp(mbox + len - 3, "/..") == 0))
+    return NULL;
+
   mutt_account_tourl (&idata->conn->account, &url);
   url.path = mbox;
   url_ciss_tostring (&url, cachepath, sizeof (cachepath), U_PATH);
@@ -130,7 +137,7 @@ HEADER* imap_hcache_get (IMAP_DATA* idata, unsigned int uid)
       h = mutt_hcache_restore ((unsigned char*)uv, NULL);
     else
       dprint (3, (debugfile, "hcache uidvalidity mismatch: %u", *uv));
-    FREE (&uv);
+    mutt_hcache_free ((void **)&uv);
   }
 
   return h;
@@ -195,7 +202,7 @@ int imap_parse_path (const char* path, IMAP_MBOX* mx)
   /* Defaults */
   memset(&mx->account, 0, sizeof(mx->account));
   mx->account.port = ImapPort;
-  mx->account.type = M_ACCT_TYPE_IMAP;
+  mx->account.type = MUTT_ACCT_TYPE_IMAP;
 
   c = safe_strdup (path);
   url_parse_ciss (&url, c);
@@ -210,7 +217,7 @@ int imap_parse_path (const char* path, IMAP_MBOX* mx)
     mx->mbox = safe_strdup (url.path);
 
     if (url.scheme == U_IMAPS)
-      mx->account.flags |= M_ACCT_SSL;
+      mx->account.flags |= MUTT_ACCT_SSL;
 
     FREE (&c);
   }
@@ -233,7 +240,7 @@ int imap_parse_path (const char* path, IMAP_MBOX* mx)
       *c = '\0';
       strfcpy (mx->account.user, tmp, sizeof (mx->account.user));
       strfcpy (tmp, c+1, sizeof (tmp));
-      mx->account.flags |= M_ACCT_USER;
+      mx->account.flags |= MUTT_ACCT_USER;
     }
 
     if ((n = sscanf (tmp, "%127[^:/]%127s", mx->account.host, tmp)) < 1)
@@ -245,11 +252,11 @@ int imap_parse_path (const char* path, IMAP_MBOX* mx)
 
     if (n > 1) {
       if (sscanf (tmp, ":%hu%127s", &(mx->account.port), tmp) >= 1)
-	mx->account.flags |= M_ACCT_PORT;
+	mx->account.flags |= MUTT_ACCT_PORT;
       if (sscanf (tmp, "/%s", tmp) == 1)
       {
 	if (!ascii_strncmp (tmp, "ssl", 3))
-	  mx->account.flags |= M_ACCT_SSL;
+	  mx->account.flags |= MUTT_ACCT_SSL;
 	else
 	{
 	  dprint (1, (debugfile, "imap_parse_path: Unknown connection type in %s\n", path));
@@ -260,7 +267,7 @@ int imap_parse_path (const char* path, IMAP_MBOX* mx)
     }
   }
 
-  if ((mx->account.flags & M_ACCT_SSL) && !(mx->account.flags & M_ACCT_PORT))
+  if ((mx->account.flags & MUTT_ACCT_SSL) && !(mx->account.flags & MUTT_ACCT_PORT))
     mx->account.port = ImapsPort;
 
   return 0;
@@ -476,7 +483,7 @@ void imap_cachepath(IMAP_DATA* idata, const char* mailbox, char* dest,
 
 /* imap_get_literal_count: write number of bytes in an IMAP literal into
  *   bytes, return 0 on success, -1 on failure. */
-int imap_get_literal_count(const char *buf, long *bytes)
+int imap_get_literal_count(const char *buf, unsigned int *bytes)
 {
   char *pc;
   char *pn;
@@ -489,7 +496,8 @@ int imap_get_literal_count(const char *buf, long *bytes)
   while (isdigit ((unsigned char) *pc))
     pc++;
   *pc = 0;
-  *bytes = atoi(pn);
+  if (mutt_atoui (pn, bytes) < 0)
+    return -1;
 
   return 0;
 }
@@ -607,28 +615,35 @@ void imap_qualify_path (char *dest, size_t len, IMAP_MBOX *mx, char* path)
 }
 
 
-/* imap_quote_string: quote string according to IMAP rules:
- *   surround string with quotes, escape " and \ with \ */
-void imap_quote_string (char *dest, size_t dlen, const char *src)
+static void _imap_quote_string (char *dest, size_t dlen, const char *src,
+                                const char *to_quote)
 {
-  static const char quote[] = "\"\\";
   char *pt;
   const char *s;
 
+  if (!(dest && dlen && src && to_quote))
+    return;
+
+  if (dlen < 3)
+  {
+    *dest = 0;
+    return;
+  }
+
   pt = dest;
   s  = src;
 
-  *pt++ = '"';
-  /* save room for trailing quote-char */
-  dlen -= 2;
+  /* save room for pre/post quote-char and trailing null */
+  dlen -= 3;
 
+  *pt++ = '"';
   for (; *s && dlen; s++)
   {
-    if (strchr (quote, *s))
+    if (strchr (to_quote, *s))
     {
+      if (dlen < 2)
+        break;
       dlen -= 2;
-      if (!dlen)
-	break;
       *pt++ = '\\';
       *pt++ = *s;
     }
@@ -642,6 +657,23 @@ void imap_quote_string (char *dest, size_t dlen, const char *src)
   *pt = 0;
 }
 
+/* imap_quote_string: quote string according to IMAP rules:
+ *   surround string with quotes, escape " and \ with \ */
+void imap_quote_string (char *dest, size_t dlen, const char *src)
+{
+  _imap_quote_string (dest, dlen, src, "\"\\");
+}
+
+/* imap_quote_string_and_backquotes: quote string according to IMAP rules:
+ *   surround string with quotes, escape " and \ with \.
+ * Additionally, escape backquotes with \ to protect against code injection
+ * when using the resulting string in mutt_parse_rc_line().
+ */
+void imap_quote_string_and_backquotes (char *dest, size_t dlen, const char *src)
+{
+  _imap_quote_string (dest, dlen, src, "\"\\`");
+}
+
 /* imap_unquote_string: equally stupid unquoting routine */
 void imap_unquote_string (char *s)
 {
@@ -748,7 +780,7 @@ void imap_keepalive (void)
   conn = mutt_socket_head ();
   while (conn)
   {
-    if (conn->account.type == M_ACCT_TYPE_IMAP)
+    if (conn->account.type == MUTT_ACCT_TYPE_IMAP)
     {
       int need_free = 0;
 
@@ -830,20 +862,20 @@ int imap_wait_keepalive (pid_t pid)
 
 void imap_allow_reopen (CONTEXT *ctx)
 {
-  if (ctx && ctx->magic == M_IMAP && CTX_DATA->ctx == ctx)
+  if (ctx && ctx->magic == MUTT_IMAP && CTX_DATA->ctx == ctx)
     CTX_DATA->reopen |= IMAP_REOPEN_ALLOW;
 }
 
 void imap_disallow_reopen (CONTEXT *ctx)
 {
-  if (ctx && ctx->magic == M_IMAP && CTX_DATA->ctx == ctx)
+  if (ctx && ctx->magic == MUTT_IMAP && CTX_DATA->ctx == ctx)
     CTX_DATA->reopen &= ~IMAP_REOPEN_ALLOW;
 }
 
 int imap_account_match (const ACCOUNT* a1, const ACCOUNT* a2)
 {
-  IMAP_DATA* a1_idata = imap_conn_find (a1, M_IMAP_CONN_NONEW);
-  IMAP_DATA* a2_idata = imap_conn_find (a2, M_IMAP_CONN_NONEW);
+  IMAP_DATA* a1_idata = imap_conn_find (a1, MUTT_IMAP_CONN_NONEW);
+  IMAP_DATA* a2_idata = imap_conn_find (a2, MUTT_IMAP_CONN_NONEW);
   const ACCOUNT* a1_canon = a1_idata == NULL ? a1 : &a1_idata->conn->account;
   const ACCOUNT* a2_canon = a2_idata == NULL ? a2 : &a2_idata->conn->account;
 
