@@ -31,6 +31,9 @@
 #include "url.h"
 #include "mutt_crypt.h"
 #include "mutt_idna.h"
+#ifdef USE_SIDEBAR
+#include "sidebar.h"
+#endif
 
 #ifdef USE_SASL
 #include "mutt_sasl.h"
@@ -64,7 +67,8 @@
 
 static const char *ReachingUs = N_("\
 To contact the developers, please mail to <mutt-dev@mutt.org>.\n\
-To report a bug, please visit http://bugs.mutt.org/.\n");
+To report a bug, please contact the Mutt maintainers via gitlab:\n\
+    https://gitlab.com/muttmua/mutt/issues\n");
 
 static const char *Notice = N_("\
 Copyright (C) 1996-2016 Michael R. Elkins and others.\n\
@@ -72,17 +76,18 @@ Mutt comes with ABSOLUTELY NO WARRANTY; for details type `mutt -vv'.\n\
 Mutt is free software, and you are welcome to redistribute it\n\
 under certain conditions; type `mutt -vv' for details.\n");
 
-static const char *Copyright = N_("\
-Copyright (C) 1996-2014 Michael R. Elkins <me@mutt.org>\n\
+static const char Copyright[] = "\
+Copyright (C) 1996-2016 Michael R. Elkins <me@mutt.org>\n\
 Copyright (C) 1996-2002 Brandon Long <blong@fiction.net>\n\
 Copyright (C) 1997-2009 Thomas Roessler <roessler@does-not-exist.org>\n\
 Copyright (C) 1998-2005 Werner Koch <wk@isil.d.shuttle.de>\n\
-Copyright (C) 1999-2014 Brendan Cully <brendan@kublai.com>\n\
+Copyright (C) 1999-2017 Brendan Cully <brendan@kublai.com>\n\
 Copyright (C) 1999-2002 Tommi Komulainen <Tommi.Komulainen@iki.fi>\n\
 Copyright (C) 2000-2004 Edmund Grimley Evans <edmundo@rano.org>\n\
 Copyright (C) 2006-2009 Rocco Rutte <pdmef@gmx.net>\n\
-Copyright (C) 2014-2015 Kevin J. McCarthy <kevin@8t8.us>\n\
-\n\
+Copyright (C) 2014-2018 Kevin J. McCarthy <kevin@8t8.us>\n";
+
+static const char *Thanks = N_("\
 Many others not mentioned here contributed code, fixes,\n\
 and suggestions.\n");
 
@@ -102,6 +107,8 @@ static const char *Obtaining = N_("\
     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n\
 ");
 
+char **envlist;
+
 void mutt_exit (int code)
 {
   mutt_endwin (NULL);
@@ -158,9 +165,9 @@ options:\n\
   exit (0);
 }
 
-extern const char cc_version[];
-extern const char cc_cflags[];
-extern const char configure_options[];
+extern unsigned char cc_version[];
+extern unsigned char cc_cflags[];
+extern unsigned char configure_options[];
 
 static char *
 rstrip_in_place(char *s)
@@ -211,19 +218,24 @@ static void show_version (void)
 	  STRINGPREP_VERSION);
 #endif
 
+#ifdef HAVE_LIBIDN2
+  printf ("\nlibidn2: %s (compiled with %s)", idn2_check_version (NULL),
+	  IDN2_VERSION);
+#endif
+
 #ifdef USE_HCACHE
   printf ("\nhcache backend: %s", mutt_hcache_backend ());
 #endif
 
   puts ("\n\nCompiler:");
   rstrip_in_place((char *)cc_version);
-  puts (cc_version);
+  puts ((char *)cc_version);
 
   rstrip_in_place((char *)configure_options);
-  printf ("\nConfigure options: %s\n", configure_options);
+  printf ("\nConfigure options: %s\n", (char *)configure_options);
 
   rstrip_in_place((char *)cc_cflags);
-  printf ("\nCompilation CFLAGS: %s\n", cc_cflags);
+  printf ("\nCompilation CFLAGS: %s\n", (char *)cc_cflags);
 
   puts (_("\nCompile options:"));
 
@@ -387,6 +399,12 @@ static void show_version (void)
 #else
 	"-HAVE_RESIZETERM  "
 #endif
+	
+#ifdef HAVE_FUTIMENS
+	"+HAVE_FUTIMENS  "
+#else
+	"-HAVE_FUTIMENS  "
+#endif
         );	
   
   puts (
@@ -472,7 +490,13 @@ static void show_version (void)
 #else
 	"-HAVE_LIBIDN  "
 #endif
-	
+
+#if HAVE_LIBIDN2
+	"+HAVE_LIBIDN2  "
+#else
+	"-HAVE_LIBIDN2  "
+#endif
+
 #if HAVE_GETSID
 	"+HAVE_GETSID  "
 #else
@@ -485,6 +509,18 @@ static void show_version (void)
 	"-USE_HCACHE  "
 #endif
 
+#ifdef USE_SIDEBAR
+	"+USE_SIDEBAR  "
+#else
+	"-USE_SIDEBAR  "
+#endif
+
+#ifdef USE_COMPRESSED
+	"+USE_COMPRESSED  "
+#else
+	"-USE_COMPRESSED  "
+#endif
+
 	);
 
 #ifdef ISPELL
@@ -504,6 +540,8 @@ static void show_version (void)
   puts ("-MIXMASTER");
 #endif
 
+  putchar ('\n');
+
   puts(_(ReachingUs));
 
   mutt_print_patchlist();
@@ -547,15 +585,16 @@ static void start_curses (void)
   meta (stdscr, TRUE);
 #endif
 init_extended_keys();
+  mutt_reflow_windows ();
 }
 
-#define M_IGNORE  (1<<0)	/* -z */
-#define M_BUFFY   (1<<1)	/* -Z */
-#define M_NOSYSRC (1<<2)	/* -n */
-#define M_RO      (1<<3)	/* -R */
-#define M_SELECT  (1<<4)	/* -y */
+#define MUTT_IGNORE  (1<<0)	/* -z */
+#define MUTT_BUFFY   (1<<1)	/* -Z */
+#define MUTT_NOSYSRC (1<<2)	/* -n */
+#define MUTT_RO      (1<<3)	/* -R */
+#define MUTT_SELECT  (1<<4)	/* -y */
 
-int main (int argc, char **argv)
+int main (int argc, char **argv, char **environ)
 {
   char folder[_POSIX_PATH_MAX] = "";
   char *subject = NULL;
@@ -587,14 +626,19 @@ int main (int argc, char **argv)
     exit(1);
   }
 
-#ifdef ENABLE_NLS
-  /* FIXME what about init.c:1439 ? */
   setlocale (LC_ALL, "");
-  bindtextdomain (PACKAGE, MUTTLOCALEDIR);
-  textdomain (PACKAGE);
-#endif
 
-  setlocale (LC_CTYPE, "");
+#ifdef ENABLE_NLS
+  /* FIXME what about the LOCALES_HACK in mutt_init() [init.c] ? */
+  {
+    char *domdir = getenv ("TEXTDOMAINDIR");
+    if (domdir && domdir[0])
+      bindtextdomain (PACKAGE, domdir);
+    else
+      bindtextdomain (PACKAGE, MUTTLOCALEDIR);
+    textdomain (PACKAGE);
+  }
+#endif
 
   mutt_error = mutt_nocurses_error;
   mutt_message = mutt_nocurses_error;
@@ -604,6 +648,17 @@ int main (int argc, char **argv)
   memset (Options, 0, sizeof (Options));
   memset (QuadOptions, 0, sizeof (QuadOptions));
 
+  /* Init envlist */
+  {
+    char **srcp, **dstp;
+    int count = 0;
+    for (srcp = environ; srcp && *srcp; srcp++)
+      count++;
+    envlist = safe_calloc(count+1, sizeof(char *));
+    for (srcp = environ, dstp = envlist; srcp && *srcp; srcp++, dstp++)
+      *dstp = safe_strdup(*srcp);
+  }
+
   for (optind = 1; optind < double_dash; )
   {
     /* We're getopt'ing POSIXLY, so we'll be here every time getopt()
@@ -671,7 +726,7 @@ int main (int argc, char **argv)
 	}
 	printf (_("Debugging at level %d.\n"), debuglevel);
 #else
-	printf _("DEBUG was not defined during compilation.  Ignored.\n");
+	printf ("%s", _("DEBUG was not defined during compilation.  Ignored.\n"));
 #endif
 	break;
 
@@ -697,7 +752,7 @@ int main (int argc, char **argv)
 	break;
 
       case 'n':
-	flags |= M_NOSYSRC;
+	flags |= MUTT_NOSYSRC;
 	break;
 
       case 'p':
@@ -709,7 +764,7 @@ int main (int argc, char **argv)
         break;
       
       case 'R':
-	flags |= M_RO; /* read-only mode */
+	flags |= MUTT_RO; /* read-only mode */
 	break;
 
       case 's':
@@ -725,15 +780,15 @@ int main (int argc, char **argv)
 	break;
 
       case 'y': /* My special hack mode */
-	flags |= M_SELECT;
+	flags |= MUTT_SELECT;
 	break;
 
       case 'z':
-	flags |= M_IGNORE;
+	flags |= MUTT_IGNORE;
 	break;
 
       case 'Z':
-	flags |= M_BUFFY | M_IGNORE;
+	flags |= MUTT_BUFFY | MUTT_IGNORE;
 	break;
 
       default:
@@ -756,7 +811,8 @@ int main (int argc, char **argv)
       break;
     default:
       puts (mutt_make_version ());
-      puts (_(Copyright));
+      puts (Copyright);
+      puts (_(Thanks));
       puts (_(Licence));
       puts (_(Obtaining));
       puts (_(ReachingUs));
@@ -770,6 +826,10 @@ int main (int argc, char **argv)
     sendflags = SENDBATCH;
   }
 
+  /* Always create the mutt_windows because batch mode has some shared code
+   * paths that end up referencing them. */
+  mutt_init_windows ();
+
   /* This must come before mutt_init() because curses needs to be started
      before calling the init_pair() function to set the color scheme.  */
   if (!option (OPTNOCURSES))
@@ -781,7 +841,7 @@ int main (int argc, char **argv)
   }
 
   /* set defaults and read init files */
-  mutt_init (flags & M_NOSYSRC, commands);
+  mutt_init (flags & MUTT_NOSYSRC, commands);
   mutt_free_list (&commands);
 
   /* Initialize crypto backends.  */
@@ -846,7 +906,7 @@ int main (int argc, char **argv)
     if (stat (fpath, &sb) == -1 && errno == ENOENT)
     {
       snprintf (msg, sizeof (msg), _("%s does not exist. Create it?"), Maildir);
-      if (mutt_yesorno (msg, M_YES) == M_YES)
+      if (mutt_yesorno (msg, MUTT_YES) == MUTT_YES)
       {
 	if (mkdir (fpath, 0700) == -1 && errno != EEXIST)
 	  mutt_error ( _("Can't create %s: %s."), Maildir, strerror (errno));
@@ -859,6 +919,7 @@ int main (int argc, char **argv)
     if (!option (OPTNOCURSES))
       mutt_flushinp ();
     ci_send_message (SENDPOSTPONED, NULL, NULL, NULL, NULL);
+    mutt_free_windows ();
     mutt_endwin (NULL);
   }
   else if (subject || msg || sendflags || draftFile || includeFile || attach ||
@@ -1147,6 +1208,7 @@ int main (int argc, char **argv)
       FREE (&tempfile);
     }
 
+    mutt_free_windows ();
     if (!option (OPTNOCURSES))
       mutt_endwin (NULL);
 
@@ -1155,7 +1217,7 @@ int main (int argc, char **argv)
   }
   else
   {
-    if (flags & M_BUFFY)
+    if (flags & MUTT_BUFFY)
     {
       if (!mutt_buffy_check (0))
       {
@@ -1165,14 +1227,14 @@ int main (int argc, char **argv)
       folder[0] = 0;
       mutt_buffy (folder, sizeof (folder));
     }
-    else if (flags & M_SELECT)
+    else if (flags & MUTT_SELECT)
     {
       if (!Incoming) {
 	mutt_endwin _("No incoming mailboxes defined.");
 	exit (1);
       }
       folder[0] = 0;
-      mutt_select_file (folder, sizeof (folder), M_SEL_FOLDER | M_SEL_BUFFY);
+      mutt_select_file (folder, sizeof (folder), MUTT_SEL_FOLDER | MUTT_SEL_BUFFY);
       if (!folder[0])
       {
 	mutt_endwin (NULL);
@@ -1187,7 +1249,7 @@ int main (int argc, char **argv)
     mutt_str_replace (&CurrentFolder, folder);
     mutt_str_replace (&LastFolder, folder);
 
-    if (flags & M_IGNORE)
+    if (flags & MUTT_IGNORE)
     {
       /* check to see if there are any messages in the folder */
       switch (mx_check_empty (folder))
@@ -1203,9 +1265,12 @@ int main (int argc, char **argv)
 
     mutt_folder_hook (folder);
 
-    if((Context = mx_open_mailbox (folder, ((flags & M_RO) || option (OPTREADONLY)) ? M_READONLY : 0, NULL))
+    if((Context = mx_open_mailbox (folder, ((flags & MUTT_RO) || option (OPTREADONLY)) ? MUTT_READONLY : 0, NULL))
        || !explicit_folder)
     {
+#ifdef USE_SIDEBAR
+      mutt_sb_set_open_buffy ();
+#endif
       mutt_index_menu ();
       if (Context)
 	FREE (&Context);
@@ -1217,6 +1282,7 @@ int main (int argc, char **argv)
     mutt_sasl_done ();
 #endif
     mutt_free_opts ();
+    mutt_free_windows ();
     mutt_endwin (Errorbuf);
   }
 
