@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-7,2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 1996-1997,2000,2010 Michael R. Elkins <me@mutt.org>
  * Copyright (C) 1998-2005 Thomas Roessler <roessler@does-not-exist.org>
  * Copyright (C) 2004 g10 Code GmbH
  *
@@ -31,6 +31,7 @@
 #endif
 
 #include "mutt.h"
+#include "mutt_crypt.h"
 #include "mutt_curses.h"
 #include "pgp.h"
 #include "mime.h"
@@ -108,20 +109,45 @@ int pgp_use_gpg_agent (void)
 {
   char *tty;
 
-  if (!option (OPTUSEGPGAGENT) || !getenv ("GPG_AGENT_INFO"))
+  /* GnuPG 2.1 no longer exports GPG_AGENT_INFO */
+  if (!option (OPTUSEGPGAGENT))
     return 0;
 
   if ((tty = ttyname(0)))
+  {
     setenv("GPG_TTY", tty, 0);
+    mutt_envlist_set ("GPG_TTY", tty, 0);
+  }
 
   return 1;
 }
 
-char *pgp_keyid(pgp_key_t k)
+static pgp_key_t _pgp_parent(pgp_key_t k)
 {
   if((k->flags & KEYFLAG_SUBKEY) && k->parent && option(OPTPGPIGNORESUB))
     k = k->parent;
 
+  return k;
+}
+
+char *pgp_long_keyid(pgp_key_t k)
+{
+  k = _pgp_parent(k);
+
+  return k->keyid;
+}
+
+char *pgp_short_keyid(pgp_key_t k)
+{
+  k = _pgp_parent(k);
+
+  return k->keyid + 8;
+}
+
+char *pgp_keyid(pgp_key_t k)
+{
+  k = _pgp_parent(k);
+
   return _pgp_keyid(k);
 }
 
@@ -133,6 +159,27 @@ char *_pgp_keyid(pgp_key_t k)
     return (k->keyid + 8);
 }
 
+char *pgp_fingerprint(pgp_key_t k)
+{
+  k = _pgp_parent(k);
+
+  return k->fingerprint;
+}
+
+/* Grab the longest key identifier available: fingerprint or else
+ * the long keyid.
+ *
+ * The longest available should be used for internally identifying
+ * the key and for invoking pgp commands.
+ */
+char *pgp_fpr_or_lkeyid(pgp_key_t k)
+{
+  char *fingerprint;
+
+  fingerprint = pgp_fingerprint (k);
+  return fingerprint ? fingerprint : pgp_long_keyid (k);
+}
+
 /* ----------------------------------------------------------------------------
  * Routines for handing PGP input.
  */
@@ -180,6 +227,113 @@ static int pgp_copy_checksig (FILE *fpin, FILE *fpout)
   return rv;
 }
 
+/* Checks PGP output messages to look for the $pgp_decryption_okay message.
+ * This protects against messages with multipart/encrypted headers
+ * but which aren't actually encrypted.  See ticket #3770
+ */
+static int pgp_check_pgp_decryption_okay_regexp (FILE *fpin)
+{
+  int rv = -1;
+
+  if (PgpDecryptionOkay.pattern)
+  {
+    char *line = NULL;
+    int lineno = 0;
+    size_t linelen;
+
+    while ((line = mutt_read_line (line, &linelen, fpin, &lineno, 0)) != NULL)
+    {
+      if (regexec (PgpDecryptionOkay.rx, line, 0, NULL, 0) == 0)
+      {
+        dprint (2, (debugfile, "pgp_check_pgp_decryption_okay_regexp: \"%s\" matches regexp.\n",
+                    line));
+        rv = 0;
+        break;
+      }
+      else
+        dprint (2, (debugfile, "pgp_check_pgp_decryption_okay_regexp: \"%s\" doesn't match regexp.\n",
+                    line));
+    }
+    FREE (&line);
+  }
+  else
+  {
+    dprint (2, (debugfile, "pgp_check_pgp_decryption_okay_regexp: No pattern.\n"));
+    rv = 1;
+  }
+
+  return rv;
+}
+
+/* Checks GnuPGP status fd output for various status codes indicating
+ * an issue.  If $pgp_check_gpg_decrypt_status_fd is unset, it falls
+ * back to the old behavior of just scanning for $pgp_decryption_okay.
+ *
+ * pgp_decrypt_part() should fail if the part is not encrypted, so we return
+ * less than 0 to indicate part or all was NOT actually encrypted.
+ *
+ * On the other hand, for pgp_application_pgp_handler(), a
+ * "BEGIN PGP MESSAGE" could indicate a signed and armored message.
+ * For that we allow -1 and -2 as "valid" (with a warning).
+ *
+ * Returns:
+ *   1 - no patterns were matched (if delegated to decryption_okay_regexp)
+ *   0 - DECRYPTION_OKAY was seen, with no PLAINTEXT outside.
+ *  -1 - No decryption status codes were encountered
+ *  -2 - PLAINTEXT was encountered outside of DECRYPTION delimeters.
+ *  -3 - DECRYPTION_FAILED was encountered
+ */
+static int pgp_check_decryption_okay (FILE *fpin)
+{
+  int rv = -1;
+  char *line = NULL, *s;
+  int lineno = 0;
+  size_t linelen;
+  int inside_decrypt = 0;
+
+  if (!option (OPTPGPCHECKGPGDECRYPTSTATUSFD))
+    return pgp_check_pgp_decryption_okay_regexp (fpin);
+
+  while ((line = mutt_read_line (line, &linelen, fpin, &lineno, 0)) != NULL)
+  {
+    if (strncmp (line, "[GNUPG:] ", 9) != 0)
+      continue;
+    s = line + 9;
+    dprint (2, (debugfile, "pgp_check_decryption_okay: checking \"%s\".\n",
+                  line));
+    if (mutt_strncmp (s, "BEGIN_DECRYPTION", 16) == 0)
+      inside_decrypt = 1;
+    else if (mutt_strncmp (s, "END_DECRYPTION", 14) == 0)
+      inside_decrypt = 0;
+    else if (mutt_strncmp (s, "PLAINTEXT", 9) == 0)
+    {
+      if (!inside_decrypt)
+      {
+        dprint (2, (debugfile, "\tPLAINTEXT encountered outside of DECRYPTION.\n"));
+        if (rv > -2)
+          rv = -2;
+      }
+    }
+    else if (mutt_strncmp (s, "DECRYPTION_FAILED", 17) == 0)
+    {
+      dprint (2, (debugfile, "\tDECRYPTION_FAILED encountered.  Failure.\n"));
+      rv = -3;
+      break;
+    }
+    else if (mutt_strncmp (s, "DECRYPTION_OKAY", 15) == 0)
+    {
+      /* Don't break out because we still have to check for
+       * PLAINTEXT outside of the decryption boundaries. */
+      dprint (2, (debugfile, "\tDECRYPTION_OKAY encountered.\n"));
+      if (rv > -2)
+        rv = 0;
+    }
+  }
+  FREE (&line);
+
+  return rv;
+}
+
 /* 
  * Copy a clearsigned message, and strip the signature and PGP's
  * dash-escaping.
@@ -204,9 +358,9 @@ static void pgp_copy_clearsigned (FILE *fpin, STATE *s, char *charset)
 
   /* fromcode comes from the MIME Content-Type charset label. It might
    * be a wrong label, so we want the ability to do corrections via
-   * charset-hooks. Therefore we set flags to M_ICONV_HOOK_FROM.
+   * charset-hooks. Therefore we set flags to MUTT_ICONV_HOOK_FROM.
    */
-  fc = fgetconv_open (fpin, charset, Charset, M_ICONV_HOOK_FROM);
+  fc = fgetconv_open (fpin, charset, Charset, MUTT_ICONV_HOOK_FROM);
   
   for (complete = 1, armor_header = 1;
        fgetconvs (buf, sizeof (buf), fc) != NULL;
@@ -247,14 +401,15 @@ static void pgp_copy_clearsigned (FILE *fpin, STATE *s, char *charset)
 
 int pgp_application_pgp_handler (BODY *m, STATE *s)
 {
-  int could_not_decrypt = 0;
+  int could_not_decrypt, decrypt_okay_rc;
   int needpass = -1, pgp_keyblock = 0;
   int clearsign = 0, rv, rc;
   int c = 1; /* silence GCC warning */
   long bytes;
   LOFF_T last_pos, offset;
   char buf[HUGE_STRING];
-  char outfile[_POSIX_PATH_MAX];
+  char pgpoutfile[_POSIX_PATH_MAX];
+  char pgperrfile[_POSIX_PATH_MAX];
   char tmpfname[_POSIX_PATH_MAX];
   FILE *pgpout = NULL, *pgpin = NULL, *pgperr = NULL;
   FILE *tmpfp = NULL;
@@ -267,7 +422,7 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
   char body_charset[STRING];
   mutt_get_body_charset (body_charset, sizeof (body_charset), m);
 
-  rc = 0;	/* silence false compiler warning if (s->flags & M_DISPLAY) */
+  rc = 0;	/* silence false compiler warning if (s->flags & MUTT_DISPLAY) */
 
   fseeko (s->fpin, m->offset, 0);
   last_pos = m->offset;
@@ -284,6 +439,8 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
     if (mutt_strncmp ("-----BEGIN PGP ", buf, 15) == 0)
     {
       clearsign = 0;
+      could_not_decrypt = 0;
+      decrypt_okay_rc = 0;
 
       if (mutt_strcmp ("MESSAGE-----\n", buf + 15) == 0)
         needpass = 1;
@@ -306,7 +463,7 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
 	continue;
       }
 
-      have_any_sigs = have_any_sigs || (clearsign && (s->flags & M_VERIFY));
+      have_any_sigs = have_any_sigs || (clearsign && (s->flags & MUTT_VERIFY));
 
       /* Copy PGP material to temporary file */
       mutt_mktemp (tmpfname, sizeof (tmpfname));
@@ -346,23 +503,33 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
       fflush (tmpfp);
 
       /* Invoke PGP if needed */
-      if (!clearsign || (s->flags & M_VERIFY))
+      if (!clearsign || (s->flags & MUTT_VERIFY))
       {
-	mutt_mktemp (outfile, sizeof (outfile));
-	if ((pgpout = safe_fopen (outfile, "w+")) == NULL)
+	mutt_mktemp (pgpoutfile, sizeof (pgpoutfile));
+	if ((pgpout = safe_fopen (pgpoutfile, "w+")) == NULL)
 	{
-	  mutt_perror (tmpfname);
-	  return -1;
+	  mutt_perror (pgpoutfile);
+          rc = -1;
+          goto out;
 	}
-	
-	if ((thepid = pgp_invoke_decode (&pgpin, NULL, &pgperr, -1,
-					 fileno (pgpout), -1, tmpfname,
-					 needpass)) == -1)
+        unlink (pgpoutfile);
+
+        mutt_mktemp (pgperrfile, sizeof (pgperrfile));
+        if ((pgperr = safe_fopen (pgperrfile, "w+")) == NULL)
+        {
+          mutt_perror (pgperrfile);
+          rc = -1;
+          goto out;
+        }
+        unlink (pgperrfile);
+
+	if ((thepid = pgp_invoke_decode (&pgpin, NULL, NULL,
+                                         -1, fileno (pgpout), fileno (pgperr),
+                                         tmpfname, needpass)) == -1)
 	{
 	  safe_fclose (&pgpout);
 	  maybe_goodsig = 0;
 	  pgpin = NULL;
-	  pgperr = NULL;
 	  state_attach_puts (_("[-- Error: unable to create PGP subprocess! --]\n"), s);
 	}
 	else /* PGP started successfully */
@@ -374,20 +541,33 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
               *PgpPass = 0;
 	    fprintf (pgpin, "%s\n", PgpPass);
 	  }
-	  
+
 	  safe_fclose (&pgpin);
 
-	  if (s->flags & M_DISPLAY)
+	  rv = mutt_wait_filter (thepid);
+
+          fflush (pgperr);
+          /* If we are expecting an encrypted message, verify status fd output.
+           * Note that BEGIN PGP MESSAGE does not guarantee the content is encrypted,
+           * so we need to be more selective about the value of decrypt_okay_rc.
+           *
+           * -3 indicates we actively found a DECRYPTION_FAILED.
+           * -2 and -1 indicate part or all of the content was plaintext.
+           */
+          if (needpass)
+          {
+            rewind (pgperr);
+            decrypt_okay_rc = pgp_check_decryption_okay (pgperr);
+            if (decrypt_okay_rc <= -3)
+              safe_fclose (&pgpout);
+          }
+
+	  if (s->flags & MUTT_DISPLAY)
 	  {
+            rewind (pgperr);
 	    crypt_current_time (s, "PGP");
 	    rc = pgp_copy_checksig (pgperr, s->fpout);
-	  }
-	  
-	  safe_fclose (&pgperr);
-	  rv = mutt_wait_filter (thepid);
-	  
-	  if (s->flags & M_DISPLAY)
-	  {
+
 	    if (rc == 0) have_any_sigs = 1;
 	    /*
 	     * Sig is bad if
@@ -417,7 +597,8 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
 	  pgp_void_passphrase ();
 	}
 	
-	if (could_not_decrypt && !(s->flags & M_DISPLAY))
+	if ((could_not_decrypt || (decrypt_okay_rc <= -3)) &&
+            !(s->flags & MUTT_DISPLAY))
 	{
           mutt_error _("Could not decrypt PGP message");
 	  mutt_sleep (1);
@@ -430,7 +611,7 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
        * Now, copy cleartext to the screen.
        */
 
-      if(s->flags & M_DISPLAY)
+      if(s->flags & MUTT_DISPLAY)
       {
 	if (needpass)
 	  state_attach_puts (_("[-- BEGIN PGP MESSAGE --]\n\n"), s);
@@ -457,20 +638,31 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
 
 	rewind (pgpout);
 	state_set_prefix (s);
-	fc = fgetconv_open (pgpout, expected_charset, Charset, M_ICONV_HOOK_FROM);
+	fc = fgetconv_open (pgpout, expected_charset, Charset, MUTT_ICONV_HOOK_FROM);
 	while ((c = fgetconv (fc)) != EOF)
 	  state_prefix_putc (c, s);
 	fgetconv_close (&fc);
       }
 
-      if (s->flags & M_DISPLAY)
+      /*
+       * Multiple PGP blocks can exist, so these need to be closed and
+       * unlinked inside the loop.
+       */
+      safe_fclose (&tmpfp);
+      mutt_unlink (tmpfname);
+      safe_fclose (&pgpout);
+      safe_fclose (&pgperr);
+
+      if (s->flags & MUTT_DISPLAY)
       {
 	state_putc ('\n', s);
 	if (needpass)
         {
 	  state_attach_puts (_("[-- END PGP MESSAGE --]\n"), s);
-	  if (could_not_decrypt)
+	  if (could_not_decrypt || (decrypt_okay_rc <= -3))
 	    mutt_error _("Could not decrypt PGP message");
+	  else if (decrypt_okay_rc < 0)
+	    mutt_error _("PGP message was not encrypted.");
 	  else
 	    mutt_message _("PGP message successfully decrypted.");
         }
@@ -500,11 +692,8 @@ out:
     safe_fclose (&tmpfp);
     mutt_unlink (tmpfname);
   }
-  if (pgpout)
-  {
-    safe_fclose (&pgpout);
-    mutt_unlink (outfile);
-  }
+  safe_fclose (&pgpout);
+  safe_fclose (&pgperr);
 
   FREE(&gpgcharset);
 
@@ -517,7 +706,7 @@ out:
   return rc;
 }
 
-static int pgp_check_traditional_one_body (FILE *fp, BODY *b, int tagged_only)
+static int pgp_check_traditional_one_body (FILE *fp, BODY *b)
 {
   char tempfile[_POSIX_PATH_MAX];
   char buf[HUGE_STRING];
@@ -530,9 +719,6 @@ static int pgp_check_traditional_one_body (FILE *fp, BODY *b, int tagged_only)
   if (b->type != TYPETEXT)
     return 0;
 
-  if (tagged_only && !b->tagged)
-    return 0;
-
   mutt_mktemp (tempfile, sizeof (tempfile));
   if (mutt_decode_save_attachment (fp, b, tempfile, 0, 0) != 0)
   {
@@ -577,21 +763,24 @@ static int pgp_check_traditional_one_body (FILE *fp, BODY *b, int tagged_only)
   return 1;
 }
 
-int pgp_check_traditional (FILE *fp, BODY *b, int tagged_only)
+int pgp_check_traditional (FILE *fp, BODY *b, int just_one)
 {
   int rv = 0;
   int r;
   for (; b; b = b->next)
   {
-    if (is_multipart (b))
-      rv = pgp_check_traditional (fp, b->parts, tagged_only) || rv;
+    if (!just_one && is_multipart (b))
+      rv = pgp_check_traditional (fp, b->parts, 0) || rv;
     else if (b->type == TYPETEXT)
     {
       if ((r = mutt_is_application_pgp (b)))
 	rv = rv || r;
       else
-	rv = pgp_check_traditional_one_body (fp, b, tagged_only) || rv;
+	rv = pgp_check_traditional_one_body (fp, b) || rv;
     }
+
+    if (just_one)
+      break;
   }
 
   return rv;
@@ -665,70 +854,7 @@ int pgp_verify_one (BODY *sigbdy, STATE *s, const char *tempfile)
 }
 
 
-/* Extract pgp public keys from messages or attachments */
-
-void pgp_extract_keys_from_messages (HEADER *h)
-{
-  int i;
-  char tempfname[_POSIX_PATH_MAX];
-  FILE *fpout;
-
-  if (h)
-  {
-    mutt_parse_mime_message (Context, h);
-    if(h->security & PGPENCRYPT && !pgp_valid_passphrase ())
-      return;
-  }
-
-  mutt_mktemp (tempfname, sizeof (tempfname));
-  if (!(fpout = safe_fopen (tempfname, "w")))
-  {
-    mutt_perror (tempfname);
-    return;
-  }
-
-  set_option (OPTDONTHANDLEPGPKEYS);
-  
-  if (!h)
-  {
-    for (i = 0; i < Context->vcount; i++)
-    {
-      if (Context->hdrs[Context->v2r[i]]->tagged)
-      {
-	mutt_parse_mime_message (Context, Context->hdrs[Context->v2r[i]]);
-	if (Context->hdrs[Context->v2r[i]]->security & PGPENCRYPT
-	   && !pgp_valid_passphrase())
-	{
-	  safe_fclose (&fpout);
-	  goto bailout;
-	}
-	mutt_copy_message (fpout, Context, Context->hdrs[Context->v2r[i]], 
-			   M_CM_DECODE|M_CM_CHARCONV, 0);
-      }
-    }
-  } 
-  else
-  {
-    mutt_parse_mime_message (Context, h);
-    if (h->security & PGPENCRYPT && !pgp_valid_passphrase())
-    {
-      safe_fclose (&fpout);
-      goto bailout;
-    }
-    mutt_copy_message (fpout, Context, h, M_CM_DECODE|M_CM_CHARCONV, 0);
-  }
-      
-  safe_fclose (&fpout);
-  mutt_endwin (NULL);
-  pgp_invoke_import (tempfname);
-  mutt_any_key_to_continue (NULL);
-
-  bailout:
-  
-  mutt_unlink (tempfname);
-  unset_option (OPTDONTHANDLEPGPKEYS);
-  
-}
+/* Extract pgp public keys from attachments */
 
 static void pgp_extract_keys_from_attachment (FILE *fp, BODY *top)
 {
@@ -762,7 +888,7 @@ void pgp_extract_keys_from_attachment_list (FILE *fp, int tag, BODY *top)
 {
   if(!fp)
   {
-    mutt_error _("Internal error. Inform <roessler@does-not-exist.org>.");
+    mutt_error _("Internal error.  Please submit a bug report.");
     return;
   }
 
@@ -822,7 +948,7 @@ BODY *pgp_decrypt_part (BODY *a, STATE *s, FILE *fpout, BODY *p)
   {
     safe_fclose (&pgperr);
     unlink (pgptmpfile);
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
       state_attach_puts (_("[-- Error: could not create a PGP subprocess! --]\n\n"), s);
     return (NULL);
   }
@@ -849,10 +975,18 @@ BODY *pgp_decrypt_part (BODY *a, STATE *s, FILE *fpout, BODY *p)
   safe_fclose (&pgpout);
   rv = mutt_wait_filter (thepid);
   mutt_unlink(pgptmpfile);
-  
-  if (s->flags & M_DISPLAY)
+
+  fflush (pgperr);
+  rewind (pgperr);
+  if (pgp_check_decryption_okay (pgperr) < 0)
+  {
+    mutt_error _("Decryption failed");
+    pgp_void_passphrase ();
+    return NULL;
+  }
+
+  if (s->flags & MUTT_DISPLAY)
   {
-    fflush (pgperr);
     rewind (pgperr);
     if (pgp_copy_checksig (pgperr, s->fpout) == 0 && !rv && p)
       p->goodsig = 1;
@@ -898,72 +1032,102 @@ int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
   char tempfile[_POSIX_PATH_MAX];
   STATE s;
   BODY *p = b;
-  
-  if(!mutt_is_multipart_encrypted(b))
+  int need_decode = 0;
+  int saved_type;
+  LOFF_T saved_offset;
+  size_t saved_length;
+  FILE *decoded_fp = NULL;
+  int rv = 0;
+
+  if (mutt_is_valid_multipart_pgp_encrypted (b))
+    b = b->parts->next;
+  else if (mutt_is_malformed_multipart_pgp_encrypted (b))
+  {
+    b = b->parts->next->next;
+    need_decode = 1;
+  }
+  else
     return -1;
 
-  if(!b->parts || !b->parts->next)
-    return -1;
-  
-  b = b->parts->next;
-  
   memset (&s, 0, sizeof (s));
   s.fpin = fpin;
+
+  if (need_decode)
+  {
+    saved_type = b->type;
+    saved_offset = b->offset;
+    saved_length = b->length;
+
+    mutt_mktemp (tempfile, sizeof (tempfile));
+    if ((decoded_fp = safe_fopen (tempfile, "w+")) == NULL)
+    {
+      mutt_perror (tempfile);
+      return (-1);
+    }
+    unlink (tempfile);
+
+    fseeko (s.fpin, b->offset, 0);
+    s.fpout = decoded_fp;
+
+    mutt_decode_attachment (b, &s);
+
+    fflush (decoded_fp);
+    b->length = ftello (decoded_fp);
+    b->offset = 0;
+    rewind (decoded_fp);
+    s.fpin = decoded_fp;
+    s.fpout = 0;
+  }
+
   mutt_mktemp (tempfile, sizeof (tempfile));
   if ((*fpout = safe_fopen (tempfile, "w+")) == NULL)
   {
     mutt_perror (tempfile);
-    return (-1);
+    rv = -1;
+    goto bail;
   }
   unlink (tempfile);
 
-  *cur = pgp_decrypt_part (b, &s, *fpout, p);
-
+  if ((*cur = pgp_decrypt_part (b, &s, *fpout, p)) == NULL)
+    rv = -1;
   rewind (*fpout);
-  
-  if (!*cur)
-    return -1;
-  
-  return (0);
+
+bail:
+  if (need_decode)
+  {
+    b->type = saved_type;
+    b->length = saved_length;
+    b->offset = saved_offset;
+    safe_fclose (&decoded_fp);
+  }
+
+  return rv;
 }
 
+/*
+ * This handler is passed the application/octet-stream directly.
+ * The caller must propagate a->goodsig to its parent.
+ */
 int pgp_encrypted_handler (BODY *a, STATE *s)
 {
   char tempfile[_POSIX_PATH_MAX];
   FILE *fpout, *fpin;
   BODY *tattach;
-  BODY *p = a;
   int rc = 0;
-  
-  a = a->parts;
-  if (!a || a->type != TYPEAPPLICATION || !a->subtype || 
-      ascii_strcasecmp ("pgp-encrypted", a->subtype) != 0 ||
-      !a->next || a->next->type != TYPEAPPLICATION || !a->next->subtype ||
-      ascii_strcasecmp ("octet-stream", a->next->subtype) != 0)
-  {
-    if (s->flags & M_DISPLAY)
-      state_attach_puts (_("[-- Error: malformed PGP/MIME message! --]\n\n"), s);
-    return -1;
-  }
-
-  /*
-   * Move forward to the application/pgp-encrypted body.
-   */
-  a = a->next;
 
   mutt_mktemp (tempfile, sizeof (tempfile));
   if ((fpout = safe_fopen (tempfile, "w+")) == NULL)
   {
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
       state_attach_puts (_("[-- Error: could not create temporary file! --]\n"), s);
     return -1;
   }
 
-  if (s->flags & M_DISPLAY) crypt_current_time (s, "PGP");
+  if (s->flags & MUTT_DISPLAY) crypt_current_time (s, "PGP");
 
-  if ((tattach = pgp_decrypt_part (a, s, fpout, p)) != NULL)
+  if ((tattach = pgp_decrypt_part (a, s, fpout, a)) != NULL)
   {
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
       state_attach_puts (_("[-- The following data is PGP/MIME encrypted --]\n\n"), s);
 
     fpin = s->fpin;
@@ -979,9 +1143,9 @@ int pgp_encrypted_handler (BODY *a, STATE *s)
      */
     
     if (mutt_is_multipart_signed (tattach) && !tattach->next)
-      p->goodsig |= tattach->goodsig;
+      a->goodsig |= tattach->goodsig;
     
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
     {
       state_puts ("\n", s);
       state_attach_puts (_("[-- End of PGP/MIME encrypted data --]\n"), s);
@@ -994,6 +1158,7 @@ int pgp_encrypted_handler (BODY *a, STATE *s)
   else
   {
     mutt_error _("Could not decrypt PGP message");
+    mutt_sleep (2);
     /* void the passphrase, even if it's not necessarily the problem */
     pgp_void_passphrase ();
     rc = -1;
@@ -1127,132 +1292,129 @@ BODY *pgp_sign_message (BODY *a)
   t->disposition = DISPNONE;
   t->encoding = ENC7BIT;
   t->unlink = 1; /* ok to remove this file after sending. */
+  mutt_set_parameter ("name", "signature.asc", &t->parameter);
 
   return (a);
 }
 
-static short is_numerical_keyid (const char *s)
-{
-  /* or should we require the "0x"? */
-  if (strncmp (s, "0x", 2) == 0)
-    s += 2;
-  if (strlen (s) % 8)
-    return 0;
-  while (*s)
-    if (strchr ("0123456789ABCDEFabcdef", *s++) == NULL)
-      return 0;
-  
-  return 1;
-}
-
 /* This routine attempts to find the keyids of the recipients of a message.
  * It returns NULL if any of the keys can not be found.
+ * If oppenc_mode is true, only keys that can be determined without
+ * prompting will be used.
  */
-char *pgp_findKeys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc)
+char *pgp_findKeys (ADDRESS *adrlist, int oppenc_mode)
 {
+  LIST *crypt_hook_list, *crypt_hook = NULL;
   char *keyID, *keylist = NULL;
   size_t keylist_size = 0;
   size_t keylist_used = 0;
-  ADDRESS *tmp = NULL, *addr = NULL;
-  ADDRESS **last = &tmp;
+  ADDRESS *addr = NULL;
   ADDRESS *p, *q;
-  int i;
-  pgp_key_t k_info = NULL, key = NULL;
+  pgp_key_t k_info = NULL;
+  char buf[LONG_STRING];
+  int r;
+  int key_selected;
 
   const char *fqdn = mutt_fqdn (1);
 
-  for (i = 0; i < 3; i++) 
+  for (p = adrlist; p ; p = p->next)
   {
-    switch (i)
+    key_selected = 0;
+    crypt_hook_list = crypt_hook = mutt_crypt_hook (p);
+    do
     {
-      case 0: p = to; break;
-      case 1: p = cc; break;
-      case 2: p = bcc; break;
-      default: abort ();
-    }
-    
-    *last = rfc822_cpy_adr (p, 0);
-    while (*last)
-      last = &((*last)->next);
+      q = p;
+      k_info = NULL;
+
+      if (crypt_hook != NULL)
+      {
+        keyID = crypt_hook->data;
+        r = MUTT_YES;
+        if (! oppenc_mode && option(OPTCRYPTCONFIRMHOOK))
+        {
+          snprintf (buf, sizeof (buf), _("Use keyID = \"%s\" for %s?"), keyID, p->mailbox);
+          r = mutt_yesorno (buf, MUTT_YES);
+        }
+        if (r == MUTT_YES)
+        {
+          if (crypt_is_numerical_keyid (keyID))
+          {
+            if (strncmp (keyID, "0x", 2) == 0)
+              keyID += 2;
+            goto bypass_selection;		/* you don't see this. */
+          }
+
+          /* check for e-mail address */
+          if (strchr (keyID, '@') &&
+              (addr = rfc822_parse_adrlist (NULL, keyID)))
+          {
+            if (fqdn) rfc822_qualify (addr, fqdn);
+            q = addr;
+          }
+          else if (! oppenc_mode)
+          {
+            k_info = pgp_getkeybystr (keyID, KEYFLAG_CANENCRYPT, PGP_PUBRING);
+          }
+        }
+        else if (r == MUTT_NO)
+        {
+          if (key_selected || (crypt_hook->next != NULL))
+          {
+            crypt_hook = crypt_hook->next;
+            continue;
+          }
+        }
+        else if (r == -1)
+        {
+          FREE (&keylist);
+          rfc822_free_address (&addr);
+          mutt_free_list (&crypt_hook_list);
+          return NULL;
+        }
+      }
+
+      if (k_info == NULL)
+      {
+        pgp_invoke_getkeys (q);
+        k_info = pgp_getkeybyaddr (q, KEYFLAG_CANENCRYPT, PGP_PUBRING, oppenc_mode);
+      }
+
+      if ((k_info == NULL) && (! oppenc_mode))
+      {
+        snprintf (buf, sizeof (buf), _("Enter keyID for %s: "), q->mailbox);
+        k_info = pgp_ask_for_key (buf, q->mailbox,
+                              KEYFLAG_CANENCRYPT, PGP_PUBRING);
+      }
+
+      if (k_info == NULL)
+      {
+        FREE (&keylist);
+        rfc822_free_address (&addr);
+        mutt_free_list (&crypt_hook_list);
+        return NULL;
+      }
+
+      keyID = pgp_fpr_or_lkeyid (k_info);
+
+    bypass_selection:
+      keylist_size += mutt_strlen (keyID) + 4;
+      safe_realloc (&keylist, keylist_size);
+      sprintf (keylist + keylist_used, "%s0x%s", keylist_used ? " " : "",	/* __SPRINTF_CHECKED__ */
+              keyID);
+      keylist_used = mutt_strlen (keylist);
+
+      key_selected = 1;
+
+      pgp_free_key (&k_info);
+      rfc822_free_address (&addr);
+
+      if (crypt_hook != NULL)
+        crypt_hook = crypt_hook->next;
+
+    } while (crypt_hook != NULL);
+
+    mutt_free_list (&crypt_hook_list);
   }
-
-  if (fqdn)
-    rfc822_qualify (tmp, fqdn);
-
-  tmp = mutt_remove_duplicates (tmp);
-  
-  for (p = tmp; p ; p = p->next)
-  {
-    char buf[LONG_STRING];
-
-    q = p;
-    k_info = NULL;
-
-    if ((keyID = mutt_crypt_hook (p)) != NULL)
-    {
-      int r;
-      snprintf (buf, sizeof (buf), _("Use keyID = \"%s\" for %s?"), keyID, p->mailbox);
-      if ((r = mutt_yesorno (buf, M_YES)) == M_YES)
-      {
-	if (is_numerical_keyid (keyID))
-	{
-	  if (strncmp (keyID, "0x", 2) == 0)
-	    keyID += 2;
-	  goto bypass_selection;		/* you don't see this. */
-	}
-	
-	/* check for e-mail address */
-	if (strchr (keyID, '@') && 
-	    (addr = rfc822_parse_adrlist (NULL, keyID)))
-	{
-	  if (fqdn) rfc822_qualify (addr, fqdn);
-	  q = addr;
-	}
-	else
-	  k_info = pgp_getkeybystr (keyID, KEYFLAG_CANENCRYPT, PGP_PUBRING);
-      }
-      else if (r == -1)
-      {
-	FREE (&keylist);
-	rfc822_free_address (&tmp);
-	rfc822_free_address (&addr);
-	return NULL;
-      }
-    }
-
-    if (k_info == NULL)
-      pgp_invoke_getkeys (q);
-
-    if (k_info == NULL && (k_info = pgp_getkeybyaddr (q, KEYFLAG_CANENCRYPT, PGP_PUBRING)) == NULL)
-    {
-      snprintf (buf, sizeof (buf), _("Enter keyID for %s: "), q->mailbox);
-
-      if ((key = pgp_ask_for_key (buf, q->mailbox,
-				  KEYFLAG_CANENCRYPT, PGP_PUBRING)) == NULL)
-      {
-	FREE (&keylist);
-	rfc822_free_address (&tmp);
-	rfc822_free_address (&addr);
-	return NULL;
-      }
-    }
-    else
-      key = k_info;
-
-    keyID = pgp_keyid (key);
-    
-  bypass_selection:
-    keylist_size += mutt_strlen (keyID) + 4;
-    safe_realloc (&keylist, keylist_size);
-    sprintf (keylist + keylist_used, "%s0x%s", keylist_used ? " " : "",	/* __SPRINTF_CHECKED__ */
-	     keyID);
-    keylist_used = mutt_strlen (keylist);
-
-    pgp_free_key (&key);
-    rfc822_free_address (&addr);
-
-  }
-  rfc822_free_address (&tmp);
   return (keylist);
 }
 
@@ -1567,10 +1729,14 @@ BODY *pgp_traditional_encryptsign (BODY *a, int flags, char *keylist)
   return b;
 }
 
-int pgp_send_menu (HEADER *msg, int *redraw)
+int pgp_send_menu (HEADER *msg)
 {
+  pgp_key_t p;
+  char input_signas[SHORT_STRING];
+  char *prompt, *letters, *choices;
+  char promptbuf[LONG_STRING];
   int choice;
-  
+
   if (!(WithCrypto & APPLICATION_PGP))
     return msg->security;
 
@@ -1578,93 +1744,148 @@ int pgp_send_menu (HEADER *msg, int *redraw)
   if (option (OPTPGPAUTOINLINE) && 
       !((msg->security & APPLICATION_PGP) && (msg->security & (SIGN|ENCRYPT))))
     msg->security |= INLINE;
-  
-  /* When the message is not selected for signing or encryption, the toggle
-   * between PGP/MIME and Traditional doesn't make sense.
+
+  msg->security |= APPLICATION_PGP;
+
+  /*
+   * Opportunistic encrypt is controlling encryption.  Allow to toggle
+   * between inline and mime, but not turn encryption on or off.
+   * NOTE: "Signing" and "Clearing" only adjust the sign bit, so we have different
+   *       letter choices for those.
    */
-  if (msg->security & (ENCRYPT | SIGN))
+  if (option (OPTCRYPTOPPORTUNISTICENCRYPT) && (msg->security & OPPENCRYPT))
   {
-    char prompt[LONG_STRING];
-
-    snprintf (prompt, sizeof (prompt), 
-	_("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, %s format, or (c)lear? "),
-	(msg->security & INLINE) ? _("PGP/M(i)ME") : _("(i)nline"));
-
-    /* The keys accepted for this prompt *must* match the order in the second
-     * version in the else clause since the switch statement below depends on
-     * it.  The 'i' key is appended in this version.
-     */
-    choice = mutt_multi_choice (prompt, _("esabfci"));
+    if (msg->security & (ENCRYPT | SIGN))
+    {
+      snprintf (promptbuf, sizeof (promptbuf),
+          _("PGP (s)ign, sign (a)s, %s format, (c)lear, or (o)ppenc mode off? "),
+          (msg->security & INLINE) ? _("PGP/M(i)ME") : _("(i)nline"));
+      prompt = promptbuf;
+      /* L10N: The 'f' is from "forget it", an old undocumented synonym of
+         'clear'.  Please use a corresponding letter in your language.
+         Alternatively, you may duplicate the letter 'c' is translated to.
+         This comment also applies to the five following letter sequences. */
+      letters = _("safcoi");
+      choices = "SaFCoi";
+    }
+    else
+    {
+      prompt = _("PGP (s)ign, sign (a)s, (c)lear, or (o)ppenc mode off? ");
+      letters = _("safco");
+      choices = "SaFCo";
+    }
   }
+  /*
+   * Opportunistic encryption option is set, but is toggled off
+   * for this message.
+   */
+  else if (option (OPTCRYPTOPPORTUNISTICENCRYPT))
+  {
+    /* When the message is not selected for signing or encryption, the toggle
+    * between PGP/MIME and Traditional doesn't make sense.
+    */
+    if (msg->security & (ENCRYPT | SIGN))
+    {
+
+      snprintf (promptbuf, sizeof (promptbuf), 
+          _("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, %s format, (c)lear, or (o)ppenc mode? "),
+          (msg->security & INLINE) ? _("PGP/M(i)ME") : _("(i)nline"));
+      prompt = promptbuf;
+      letters = _("esabfcoi");
+      choices = "esabfcOi";
+    }
+    else
+    {
+      prompt = _("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, (c)lear, or (o)ppenc mode? ");
+      letters = _("esabfco");
+      choices = "esabfcO";
+    }
+  }
+  /*
+   * Opportunistic encryption is unset
+   */
   else
   {
-    /* The keys accepted *must* be a prefix of the accepted keys in the "if"
-     * clause above since the switch statement below depends on it.
-     */
-    choice = mutt_multi_choice(_("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, or (c)lear? "),
-	_("esabfc"));
+    if (msg->security & (ENCRYPT | SIGN))
+    {
+
+      snprintf (promptbuf, sizeof (promptbuf), 
+          _("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, %s format, or (c)lear? "),
+          (msg->security & INLINE) ? _("PGP/M(i)ME") : _("(i)nline"));
+      prompt = promptbuf;
+      letters = _("esabfci");
+      choices = "esabfci";
+    }
+    else
+    {
+      prompt = _("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, or (c)lear? ");
+      letters = _("esabfc");
+      choices = "esabfc";
+    }
   }
 
-  switch (choice)
+  choice = mutt_multi_choice (prompt, letters);
+  if (choice > 0)
   {
-    case 1: /* (e)ncrypt */
+    switch (choices[choice - 1])
+    {
+    case 'e': /* (e)ncrypt */
       msg->security |= ENCRYPT;
       msg->security &= ~SIGN;
       break;
 
-  case 2: /* (s)ign */
-    msg->security |= SIGN;
-    msg->security &= ~ENCRYPT;
-    break;
+    case 's': /* (s)ign */
+      msg->security &= ~ENCRYPT;
+      msg->security |= SIGN;
+      break;
 
-  case 3: /* sign (a)s */
-    {
-      pgp_key_t p;
-      char input_signas[SHORT_STRING];
+    case 'S': /* (s)ign in oppenc mode */
+      msg->security |= SIGN;
+      break;
 
+    case 'a': /* sign (a)s */
       unset_option(OPTPGPCHECKTRUST);
 
       if ((p = pgp_ask_for_key (_("Sign as: "), NULL, 0, PGP_SECRING)))
       {
-	snprintf (input_signas, sizeof (input_signas), "0x%s",
-	    pgp_keyid (p));
-	mutt_str_replace (&PgpSignAs, input_signas);
-	pgp_free_key (&p);
+        snprintf (input_signas, sizeof (input_signas), "0x%s",
+            pgp_fpr_or_lkeyid (p));
+        mutt_str_replace (&PgpSignAs, input_signas);
+        pgp_free_key (&p);
 
-	msg->security |= SIGN;
+        msg->security |= SIGN;
 
-	crypt_pgp_void_passphrase ();  /* probably need a different passphrase */
+        crypt_pgp_void_passphrase ();  /* probably need a different passphrase */
       }
-#if 0
-      else
-      {
-	msg->security &= ~SIGN;
-      }
-#endif
+      break;
 
-      *redraw = REDRAW_FULL;
-    } break;
+    case 'b': /* (b)oth */
+      msg->security |= (ENCRYPT | SIGN);
+      break;
 
-  case 4: /* (b)oth */
-    msg->security |= (ENCRYPT | SIGN);
-    break;
+    case 'f': /* (f)orget it: kept for backward compatibility. */
+    case 'c': /* (c)lear     */
+      msg->security &= ~(ENCRYPT | SIGN);
+      break;
 
-  case 5: /* (f)orget it */
-  case 6: /* (c)lear     */
-    msg->security = 0;
-    break;
+    case 'F': /* (f)orget it or (c)lear in oppenc mode */
+    case 'C':
+      msg->security &= ~SIGN;
+      break;
 
-  case 7: /* toggle (i)nline */
-    msg->security ^= INLINE;
-    break;
-  }
+    case 'O': /* oppenc mode on */
+      msg->security |= OPPENCRYPT;
+      crypt_opportunistic_encrypt (msg);
+      break;
 
-  if (msg->security)
-  {
-    if (! (msg->security & (ENCRYPT | SIGN)))
-      msg->security = 0;
-    else
-      msg->security |= APPLICATION_PGP;
+    case 'o': /* oppenc mode off */
+      msg->security &= ~OPPENCRYPT;
+      break;
+
+    case 'i': /* toggle (i)nline */
+      msg->security ^= INLINE;
+      break;
+    }
   }
 
   return (msg->security);
