@@ -129,7 +129,7 @@ int mutt_num_postponed (int force)
 
     if (access (Postponed, R_OK | F_OK) != 0)
       return (PostCount = 0);
-    if (mx_open_mailbox (Postponed, M_NOSORT | M_QUIET, &ctx) == NULL)
+    if (mx_open_mailbox (Postponed, MUTT_NOSORT | MUTT_QUIET, &ctx) == NULL)
       PostCount = 0;
     else
       PostCount = ctx.msgcount;
@@ -149,7 +149,7 @@ static void post_entry (char *s, size_t slen, MUTTMENU *menu, int entry)
   CONTEXT *ctx = (CONTEXT *) menu->data;
 
   _mutt_make_string (s, slen, NONULL (HdrFmt), ctx, ctx->hdrs[entry],
-		     M_FORMAT_ARROWCURSOR);
+		     MUTT_FORMAT_ARROWCURSOR);
 }
 
 static HEADER *select_msg (void)
@@ -165,6 +165,7 @@ static HEADER *select_msg (void)
   menu->title = _("Postponed Messages");
   menu->data = PostContext;
   menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_POST, PostponeHelp);
+  mutt_push_current_menu (menu);
 
   /* The postponed mailbox is setup to have sorting disabled, but the global
    * Sort variable may indicate something different.   Sorting has to be
@@ -178,7 +179,8 @@ static HEADER *select_msg (void)
     {
       case OP_DELETE:
       case OP_UNDELETE:
-	mutt_set_flag (PostContext, PostContext->hdrs[menu->current], M_DELETE, (i == OP_DELETE) ? 1 : 0);
+        /* should deleted draft messages be saved in the trash folder? */
+	mutt_set_flag (PostContext, PostContext->hdrs[menu->current], MUTT_DELETE, (i == OP_DELETE) ? 1 : 0);
 	PostCount = PostContext->msgcount - PostContext->deleted;
 	if (option (OPTRESOLVE) && menu->current < menu->max - 1)
 	{
@@ -187,13 +189,13 @@ static HEADER *select_msg (void)
 	  if (menu->current >= menu->top + menu->pagelen)
 	  {
 	    menu->top = menu->current;
-	    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	    menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
 	  }
 	  else
 	    menu->redraw |= REDRAW_MOTION_RESYNCH;
 	}
 	else
-	  menu->redraw = REDRAW_CURRENT;
+	  menu->redraw |= REDRAW_CURRENT;
 	break;
 
       case OP_GENERIC_SELECT_ENTRY:
@@ -208,6 +210,7 @@ static HEADER *select_msg (void)
   }
 
   Sort = orig_sort;
+  mutt_pop_current_menu (menu);
   mutt_menuDestroy (&menu);
   return (r > -1 ? PostContext->hdrs[r] : NULL);
 }
@@ -239,7 +242,7 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur, char *fcc, size
   if (!Postponed)
     return (-1);
 
-  if ((PostContext = mx_open_mailbox (Postponed, M_NOSORT, NULL)) == NULL)
+  if ((PostContext = mx_open_mailbox (Postponed, MUTT_NOSORT, NULL)) == NULL)
   {
     PostCount = 0;
     mutt_error _("No postponed messages.");
@@ -275,14 +278,15 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur, char *fcc, size
   }
 
   /* finished with this message, so delete it. */
-  mutt_set_flag (PostContext, h, M_DELETE, 1);
+  mutt_set_flag (PostContext, h, MUTT_DELETE, 1);
+  mutt_set_flag (PostContext, h, MUTT_PURGE, 1);
 
   /* update the count for the status display */
   PostCount = PostContext->msgcount - PostContext->deleted;
 
   /* avoid the "purge deleted messages" prompt */
   opt_delete = quadoption (OPT_DELETE);
-  set_quadoption (OPT_DELETE, M_YES);
+  set_quadoption (OPT_DELETE, MUTT_YES);
   mx_close_mailbox (PostContext, NULL);
   set_quadoption (OPT_DELETE, opt_delete);
 
@@ -527,15 +531,23 @@ int mutt_parse_crypt_hdr (const char *p, int set_empty_signas, int crypt_app)
   if ((WithCrypto & APPLICATION_SMIME) && (crypt_app == APPLICATION_SMIME)
       && (flags & SIGN)
       && (set_empty_signas || *sign_as))
-    mutt_str_replace (&SmimeDefaultKey, sign_as);
+    mutt_str_replace (&SmimeSignAs, sign_as);
 
   return flags;
 }
 
 
-
+/* args:
+ *     fp      If not NULL, file containing the template
+ *     ctx     If fp is NULL, the context containing the header with the template
+ *     newhdr  The template is read into this HEADER
+ *     hdr     The message to recall/resend
+ *     resend  Set if resending (as opposed to recalling a postponed msg).
+ *             Resent messages enable header weeding, and also
+ *             discard any existing Message-ID and Mail-Followup-To.
+ */
 int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
-			       short weed)
+                           short resend)
 {
   MESSAGE *msg = NULL;
   char file[_POSIX_PATH_MAX];
@@ -558,15 +570,16 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
 
   fseeko (fp, hdr->offset, 0);
   newhdr->offset = hdr->offset;
-  newhdr->env = mutt_read_rfc822_header (fp, newhdr, 1, weed);
+  /* enable header weeding for resent messages */
+  newhdr->env = mutt_read_rfc822_header (fp, newhdr, 1, resend);
   newhdr->content->length = hdr->content->length;
   mutt_parse_part (fp, newhdr->content);
 
-  /* If message_id is set, then we are resending a message and don't want
-   * message_id or mail_followup_to. Otherwise, we are resuming a
-   * postponed message, and want to keep the mail_followup_to.
+  /* If resending a message, don't keep message_id or mail_followup_to.
+   * Otherwise, we are resuming a postponed message, and want to keep those
+   * headers if they exist.
    */
-  if (newhdr->env->message_id != NULL)
+  if (resend)
   {
     FREE (&newhdr->env->message_id);
     FREE (&newhdr->env->mail_followup_to);
@@ -579,18 +592,14 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
   {
     newhdr->security |= sec_type;
     if (!crypt_valid_passphrase (sec_type))
-      goto err;
+      goto bail;
 
     mutt_message _("Decrypting message...");
     if ((crypt_pgp_decrypt_mime (fp, &bfp, newhdr->content, &b) == -1)
 	|| b == NULL)
     {
- err:
-      mx_close_message (&msg);
-      mutt_free_envelope (&newhdr->env);
-      mutt_free_body (&newhdr->content);
       mutt_error _("Decryption failed.");
-      return -1;
+      goto bail;
     }
 
     mutt_free_body (&newhdr->content);
@@ -664,7 +673,7 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
 	b->noconv = 1;
       else
       {
-	s.flags |= M_CHARCONV;
+	s.flags |= MUTT_CHARCONV;
 	b->noconv = 0;
       }
 
@@ -679,7 +688,18 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
     if ((WithCrypto & APPLICATION_PGP) &&
 	((sec_type = mutt_is_application_pgp (b)) & (ENCRYPT|SIGN)))
     {
-      mutt_body_handler (b, &s);
+      if (sec_type & ENCRYPT)
+      {
+        if (!crypt_valid_passphrase (APPLICATION_PGP))
+          goto bail;
+        mutt_message _("Decrypting message...");
+      }
+
+      if (mutt_body_handler (b, &s) < 0)
+      {
+        mutt_error _("Decryption failed.");
+        goto bail;
+      }
 
       newhdr->security |= sec_type;
 
@@ -690,7 +710,19 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
     else if ((WithCrypto & APPLICATION_SMIME) &&
              ((sec_type = mutt_is_application_smime (b)) & (ENCRYPT|SIGN)))
     {
-      mutt_body_handler (b, &s);
+      if (sec_type & ENCRYPT)
+      {
+        if (!crypt_valid_passphrase (APPLICATION_SMIME))
+          goto bail;
+        crypt_smime_getkeys (newhdr->env);
+        mutt_message _("Decrypting message...");
+      }
+
+      if (mutt_body_handler (b, &s) < 0)
+      {
+        mutt_error _("Decryption failed.");
+        goto bail;
+      }
 
       newhdr->security |= sec_type;
       b->type = TYPETEXT;
@@ -735,7 +767,7 @@ int mutt_prepare_template (FILE *fp, CONTEXT *ctx, HEADER *newhdr, HEADER *hdr,
 
   /* that's it. */
   if (bfp != fp) safe_fclose (&bfp);
-  if (msg) mx_close_message (&msg);
+  if (msg) mx_close_message (ctx, &msg);
 
   if (rv == -1)
   {
