@@ -159,6 +159,7 @@ int smime_valid_passphrase (void)
 static const char *_mutt_fmt_smime_command (char *dest,
 					    size_t destlen,
 					    size_t col,
+                                            int cols,
 					    char op,
 					    const char *src,
 					    const char *prefix,
@@ -169,7 +170,7 @@ static const char *_mutt_fmt_smime_command (char *dest,
 {
   char fmt[16];
   struct smime_command_context *cctx = (struct smime_command_context *) data;
-  int optional = (flags & M_FORMAT_OPTIONAL);
+  int optional = (flags & MUTT_FORMAT_OPTIONAL);
   
   switch (op)
   {
@@ -284,10 +285,10 @@ static const char *_mutt_fmt_smime_command (char *dest,
   }
 
   if (optional)
-    mutt_FormatString (dest, destlen, col, ifstring, _mutt_fmt_smime_command,
+    mutt_FormatString (dest, destlen, col, cols, ifstring, _mutt_fmt_smime_command,
 		       data, 0);
-  else if (flags & M_FORMAT_OPTIONAL)
-    mutt_FormatString (dest, destlen, col, elsestring, _mutt_fmt_smime_command,
+  else if (flags & MUTT_FORMAT_OPTIONAL)
+    mutt_FormatString (dest, destlen, col, cols, elsestring, _mutt_fmt_smime_command,
 		       data, 0);
 
   return (src);
@@ -298,7 +299,7 @@ static const char *_mutt_fmt_smime_command (char *dest,
 static void mutt_smime_command (char *d, size_t dlen,
 				struct smime_command_context *cctx, const char *fmt)
 {
-  mutt_FormatString (d, dlen, 0, NONULL(fmt), _mutt_fmt_smime_command,
+  mutt_FormatString (d, dlen, 0, MuttIndexWindow->cols, NONULL(fmt), _mutt_fmt_smime_command,
 		    (unsigned long) cctx, 0);
   dprint (2,(debugfile, "mutt_smime_command: %s\n", d));
 }
@@ -446,6 +447,7 @@ static smime_key_t *smime_select_key (smime_key_t *keys, char *query)
   menu->help = helpstr;
   menu->data = table;
   menu->title = title;
+  mutt_push_current_menu (menu);
   /* sorting keys might be done later - TODO */
 
   mutt_clear_error();
@@ -476,7 +478,7 @@ static smime_key_t *smime_select_key (smime_key_t *keys, char *query)
           snprintf (buf, sizeof (buf), _("%s Do you really want to use the key?"),
                     _(s));
 
-          if (mutt_yesorno (buf, M_NO) != M_YES)
+          if (mutt_yesorno (buf, MUTT_NO) != MUTT_YES)
           {
             mutt_clear_error ();
             break;
@@ -492,9 +494,9 @@ static smime_key_t *smime_select_key (smime_key_t *keys, char *query)
     }
   }
 
+  mutt_pop_current_menu (menu);
   mutt_menuDestroy (&menu);
   FREE (&table);
-  set_option (OPTNEEDREDRAW);
 
   return selected_key;
 }
@@ -765,7 +767,7 @@ smime_key_t *smime_ask_for_key(char *prompt, short abilities, short public)
   FOREVER
   {
     resp[0] = 0;
-    if (mutt_get_field (prompt, resp, sizeof (resp), M_CLEAR) != 0)
+    if (mutt_get_field (prompt, resp, sizeof (resp), MUTT_CLEAR) != 0)
       return NULL;
 
     if ((key = smime_get_key_by_str (resp, abilities, public)))
@@ -909,11 +911,11 @@ char *smime_findKeys (ADDRESS *adrlist, int oppenc_mode)
       FREE (&keylist);
       return NULL;
     }
-    
+
     keyID = key->hash;
     keylist_size += mutt_strlen (keyID) + 2;
     safe_realloc (&keylist, keylist_size);
-    sprintf (keylist + keylist_used, "%s\n", keyID);	/* __SPRINTF_CHECKED__ */
+    sprintf (keylist + keylist_used, "%s%s", keylist_used ? " " : "", keyID);	/* __SPRINTF_CHECKED__ */
     keylist_used = mutt_strlen (keylist);
 
     smime_free_key (&key);
@@ -1227,7 +1229,7 @@ void smime_invoke_import (char *infile, char *mailbox)
 
   buf[0] = '\0';
   if (option (OPTASKCERTLABEL))
-    mutt_get_field ("Label for certificate:", buf, sizeof (buf), 0);
+    mutt_get_field (_("Label for certificate: "), buf, sizeof (buf), 0);
 
   mutt_endwin (NULL);
   if ((certfile = smime_extract_certificate(infile)))
@@ -1282,7 +1284,7 @@ int smime_verify_sender(HEADER *h)
 
   if(h->security & ENCRYPT)
     mutt_copy_message (fpout, Context, h,
-		       M_CM_DECODE_CRYPT & M_CM_DECODE_SMIME,
+		       MUTT_CM_DECODE_CRYPT & MUTT_CM_DECODE_SMIME,
 		       CH_MIME|CH_WEED|CH_NONEWLINE);
   else
     mutt_copy_message (fpout, Context, h, 0, 0);
@@ -1372,10 +1374,10 @@ BODY *smime_build_smime_entity (BODY *a, char *certlist)
   char buf[LONG_STRING], certfile[LONG_STRING];
   char tempfile[_POSIX_PATH_MAX], smimeerrfile[_POSIX_PATH_MAX];
   char smimeinfile[_POSIX_PATH_MAX];
-  char *cert_start = certlist, *cert_end = certlist;
+  char *cert_start, *cert_end;
   FILE *smimein = NULL, *smimeerr = NULL, *fpout = NULL, *fptmp = NULL;
   BODY *t;
-  int err = 0, empty;
+  int err = 0, empty, off;
   pid_t thepid;
   
   mutt_mktemp (tempfile, sizeof (tempfile));
@@ -1389,8 +1391,8 @@ BODY *smime_build_smime_entity (BODY *a, char *certlist)
   if ((smimeerr = safe_fopen (smimeerrfile, "w+")) == NULL)
   {
     mutt_perror (smimeerrfile);
-    safe_fclose (&fpout);
     mutt_unlink (tempfile);
+    safe_fclose (&fpout);
     return NULL;
   }
   mutt_unlink (smimeerrfile);
@@ -1406,17 +1408,18 @@ BODY *smime_build_smime_entity (BODY *a, char *certlist)
   }
 
   *certfile = '\0';
-  while (1)
+  for (cert_start = certlist; cert_start; cert_start = cert_end)
   {
-    int off = mutt_strlen (certfile);
-    while (*++cert_end && *cert_end != '\n');
-    if (!*cert_end) break;
-    *cert_end = '\0';
-    snprintf (certfile+off, sizeof (certfile)-off, " %s/%s",
-	      NONULL(SmimeCertificates), cert_start);
-    *cert_end = '\n';
-    cert_start = cert_end;
-    cert_start++;
+    if ((cert_end = strchr (cert_start, ' ')))
+      *cert_end = '\0';
+    if (*cert_start)
+    {
+      off = mutt_strlen (certfile);
+      snprintf (certfile+off, sizeof (certfile)-off, "%s%s/%s",
+                off ? " " : "", NONULL(SmimeCertificates), cert_start);
+    }
+    if (cert_end)
+      *cert_end++ = ' ';
   }
 
   /* write a MIME entity */
@@ -1430,9 +1433,10 @@ BODY *smime_build_smime_entity (BODY *a, char *certlist)
 			     fileno (fpout), fileno (smimeerr),
 			     smimeinfile, certfile)) == -1)
   {
+    mutt_unlink (tempfile);
+    safe_fclose (&fpout);
     safe_fclose (&smimeerr);
     mutt_unlink (smimeinfile);
-    mutt_unlink (certfile);
     return (NULL);
   }
 
@@ -1440,7 +1444,6 @@ BODY *smime_build_smime_entity (BODY *a, char *certlist)
   
   mutt_wait_filter (thepid);
   mutt_unlink (smimeinfile);
-  mutt_unlink (certfile);
   
   fflush (fpout);
   rewind (fpout);
@@ -1524,11 +1527,13 @@ BODY *smime_sign_message (BODY *a )
   int err = 0;
   int empty = 0;
   pid_t thepid;
-  smime_key_t *default_key;
+  char *signas;
+  smime_key_t *signas_key;
   char *intermediates;
   char *micalg;
 
-  if (!SmimeDefaultKey)
+  signas = (SmimeSignAs && *SmimeSignAs) ? SmimeSignAs : SmimeDefaultKey;
+  if (!signas || !*signas)
   {
     mutt_error _("Can't sign: No key specified. Use Sign As.");
     return NULL;
@@ -1560,22 +1565,22 @@ BODY *smime_sign_message (BODY *a )
   
 
   snprintf (SmimeKeyToUse, sizeof (SmimeKeyToUse), "%s/%s", 
-	   NONULL(SmimeKeys), SmimeDefaultKey);
+	   NONULL(SmimeKeys), signas);
 
   snprintf (SmimeCertToUse, sizeof (SmimeCertToUse), "%s/%s",
-	   NONULL(SmimeCertificates), SmimeDefaultKey);
+	   NONULL(SmimeCertificates), signas);
   
-  default_key = smime_get_key_by_hash (SmimeDefaultKey, 1);
-  if ((! default_key) ||
-      (! mutt_strcmp ("?", default_key->issuer)))
-    intermediates = SmimeDefaultKey; /* so openssl won't complain in any case */
+  signas_key = smime_get_key_by_hash (signas, 1);
+  if ((! signas_key) ||
+      (! mutt_strcmp ("?", signas_key->issuer)))
+    intermediates = signas; /* so openssl won't complain in any case */
   else
-    intermediates = default_key->issuer;
+    intermediates = signas_key->issuer;
 
   snprintf (SmimeIntermediateToUse, sizeof (SmimeIntermediateToUse), "%s/%s",
 	   NONULL(SmimeCertificates), intermediates);
 
-  smime_free_key (&default_key);
+  smime_free_key (&signas_key);
   
 
 
@@ -1869,7 +1874,7 @@ static BODY *smime_handle_entity (BODY *m, STATE *s, FILE *outFile)
   {
     safe_fclose (&smimeout); smimeout = NULL;
     mutt_unlink (tmpfname);
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
       state_attach_puts (_("[-- Error: unable to create OpenSSL subprocess! --]\n"), s);
     return NULL;
   }
@@ -1880,7 +1885,7 @@ static BODY *smime_handle_entity (BODY *m, STATE *s, FILE *outFile)
   {
     safe_fclose (&smimeout); smimeout = NULL;
     mutt_unlink (tmpfname);
-    if (s->flags & M_DISPLAY)
+    if (s->flags & MUTT_DISPLAY)
       state_attach_puts (_("[-- Error: unable to create OpenSSL subprocess! --]\n"), s);
     return NULL;
   }
@@ -1900,7 +1905,7 @@ static BODY *smime_handle_entity (BODY *m, STATE *s, FILE *outFile)
   mutt_unlink (tmpfname);
   
 
-  if (s->flags & M_DISPLAY)
+  if (s->flags & MUTT_DISPLAY)
   {
     fflush (smimeerr);
     rewind (smimeerr);
@@ -1979,7 +1984,7 @@ static BODY *smime_handle_entity (BODY *m, STATE *s, FILE *outFile)
     fpout = NULL;
   }
 
-  if (s->flags & M_DISPLAY)
+  if (s->flags & MUTT_DISPLAY)
   {
     if (type & ENCRYPT)
       state_attach_puts (_("\n[-- End of S/MIME encrypted data. --]\n"), s);
@@ -2088,7 +2093,7 @@ int smime_application_smime_handler (BODY *m, STATE *s)
   return smime_handle_entity (m, s, NULL) ? 0 : -1;
 }
 
-int smime_send_menu (HEADER *msg, int *redraw)
+int smime_send_menu (HEADER *msg)
 {
   smime_key_t *key;
   char *prompt, *letters, *choices;
@@ -2213,21 +2218,11 @@ int smime_send_menu (HEADER *msg, int *redraw)
       break;
 
     case 's': /* (s)ign */
-    case 'S': /* (s)ign in oppenc mode */
-      if(!SmimeDefaultKey)
-      {
-        *redraw = REDRAW_FULL;
+      msg->security &= ~ENCRYPT;
+      msg->security |= SIGN;
+      break;
 
-        if ((key = smime_ask_for_key (_("Sign as: "), KEYFLAG_CANSIGN, 0)))
-        {
-          mutt_str_replace (&SmimeDefaultKey, key->hash);
-          smime_free_key (&key);
-        }
-        else
-          break;
-      }
-      if (choices[choice - 1] == 's')
-        msg->security &= ~ENCRYPT;
+    case 'S': /* (s)ign in oppenc mode */
       msg->security |= SIGN;
       break;
 
@@ -2235,16 +2230,15 @@ int smime_send_menu (HEADER *msg, int *redraw)
 
       if ((key = smime_ask_for_key (_("Sign as: "), KEYFLAG_CANSIGN, 0))) 
       {
-        mutt_str_replace (&SmimeDefaultKey, key->hash);
+        mutt_str_replace (&SmimeSignAs, key->hash);
         smime_free_key (&key);
-          
+
         msg->security |= SIGN;
 
         /* probably need a different passphrase */
         crypt_smime_void_passphrase ();
       }
 
-      *redraw = REDRAW_FULL;
       break;
 
     case 'b': /* (b)oth */
